import inspect
from typing import Any, Hashable, List


class AntidoteError(Exception):
    """ Base class of all errors of antidote. """

    def __repr__(self):
        return "{}({})".format(type(self).__name__, str(self))


class DuplicateDependencyError(AntidoteError):
    """
    A dependency already exists with the same id.
    *May* be raised by providers.
    """

    def __init__(self, dependency: Hashable, existing_definition: Any):
        self.dependency = dependency
        self.existing_definition = existing_definition

    def __repr__(self):
        return ("DuplicateDependencyError(dependency={!r}, "
                "existing_definition={!r})").format(
            self.dependency, self.existing_definition
        )

    def __str__(self):
        return "The dependency {} already exists. It points to {}.".format(
            self.dependency,
            self.existing_definition
        )


class DependencyInstantiationError(AntidoteError):
    """
    The dependency could not be instantiated.
    Raised by the core.
    """


class DependencyCycleError(AntidoteError):
    """
    A dependency cycle is found.
    Raised by the core.
    """

    def __init__(self, dependencies: List):
        self.dependencies = dependencies

    def __str__(self):
        return ' => '.join(map(self._repr, self.dependencies))

    @staticmethod
    def _repr(dependency: Hashable) -> str:
        if inspect.isclass(dependency):
            return "{}.{}".format(dependency.__module__,
                                  getattr(dependency, '__name__', dependency))

        return repr(dependency)


class DependencyNotFoundError(AntidoteError):
    """
    The dependency could not be found in the core.
    Raised by the core.
    """

    def __init__(self, dependency: Hashable):
        self.missing_dependency = dependency

    def __str__(self):
        return repr(self.missing_dependency)



import functools
from typing import Callable

import pytest

from antidote import (factory, inject, LazyConstantsMeta, new_container, provider,
                      register, wire)
from antidote.core import DependencyContainer, DependencyProvider
from antidote.exceptions import DependencyInstantiationError


class Service:
    def __repr__(self):
        return "Service"


class AnotherService:
    def __repr__(self):
        return "AnotherService"


class YetAnotherService:
    def __repr__(self):
        return "YetAnotherService"


class SuperService:
    def __repr__(self):
        return "SuperService"


@pytest.fixture()
def container():
    c = new_container()
    c.update_singletons({cls: cls() for cls in [Service, AnotherService,
                                                YetAnotherService]})
    c.update_singletons(dict(service=object(),
                             another_service=object()))
    return c


class MyService:
    def __init__(self,
                 service: Service,
                 another_service=None):
        self.service = service
        self.another_service = another_service

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


class F1:
    def __init__(self,
                 service: Service,
                 another_service=None):
        self.service = service
        self.another_service = another_service

    def __call__(self) -> MyService:
        return MyService(self.service, self.another_service)

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


class F2:
    def __init__(self, service: Service):
        self.service = service

    def __call__(self, another_service=None) -> MyService:
        return MyService(self.service, another_service)

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


class F3:
    def __init__(self):
        pass

    def __call__(self,
                 service: Service,
                 another_service=None) -> MyService:
        return MyService(service, another_service)

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


class B1:
    def __init__(self, s, a):
        self.service = s
        self.another_service = a

    @classmethod
    def build(cls,
              service: Service,
              another_service=None):
        return cls(service, another_service)

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


class MyProvider(DependencyProvider):
    def __init__(self,
                 container: DependencyContainer,
                 service: Service,
                 another_service=None):
        self._container = container
        self.service = service
        self.another_service = another_service

    def provide(self, dependency):
        return

    def method(self, yet_another_service: YetAnotherService):
        return yet_another_service


def f1(service: Service, another_service=None) -> MyService:
    return MyService(service, another_service)


def g1(x, service: Service, another_service=None) -> MyService:
    return MyService(service, another_service)


def build(cls, service: Service, another_service=None):
    return cls(service, another_service)


def wire_(class_=None, auto_wire=True, **kwargs):
    if auto_wire is True:
        m = ['__init__']
    elif auto_wire is False:
        m = []
    else:
        m = auto_wire

    return wire(class_=class_, methods=m, **kwargs)


def resource(class_, **kwargs):
    return LazyConstantsMeta(
        'Resource' + class_.__name__,
        (),
        dict(
            __init__=class_.__init__,
            __call__=class_.__call__,
            method=class_.method
        ),
        lazy_method='__call__',
        **kwargs
    )


register_build = functools.partial(register, factory='build')
register_external_build = functools.partial(register, factory=build)

class_tests = [
    pytest.param(provider, MyProvider,
                 id='provider-MyProvider'),
    pytest.param(wire_, MyService,
                 id='wire-MyService'),
    pytest.param(register, MyService,
                 id='register-MyService'),
    pytest.param(factory, F1,
                 id='factory-F1'),
    pytest.param(factory, F2,
                 id='factory-F2'),
    pytest.param(factory, F3,
                 id='factory-F3'),
    pytest.param(register_build, B1,
                 id='register_build-B1'),
]

metaclass_tests = [
    pytest.param(resource, F1,
                 id='resource-F1'),
    pytest.param(resource, F2,
                 id='resource-F2'),
    pytest.param(resource, F3,
                 id='resource-F3')
]

function_tests = [
    pytest.param(factory, f1,
                 id='factory-f1'),
    pytest.param(inject, f1,
                 id='inject-f1'),
    pytest.param(register_external_build, B1,
                 id='register_external_build-B1'),
]

all_tests = metaclass_tests + class_tests + function_tests


def parametrize_injection(tests, lazy=False, return_wrapped=False,  # noqa: C901
                          create_subclass=False,
                          **inject_kwargs):
    def decorator(test):
        @pytest.mark.parametrize('wrapper,wrapped', tests)
        def f(container, wrapper, wrapped):
            if isinstance(wrapped, type):
                if create_subclass:
                    def __init__(*args, **kwargs):
                        pass

                    # Subclass to ensure wire_super is working properly.
                    wrapped = type("Sub" + wrapped.__name__,
                                   (wrapped,),
                                   {'__init__': __init__,
                                    'build': classmethod(lambda cls: cls())})
                else:
                    # helpers do modify the class, so a copy has to be made to
                    # avoid any conflict between the tests.
                    wrapped = type(wrapped.__name__,
                                   wrapped.__bases__,
                                   wrapped.__dict__.copy())

            def create():
                inj_kwargs = inject_kwargs.copy()

                if wrapper == register_build:
                    try:
                        auto_wire = inj_kwargs['auto_wire']
                        if isinstance(auto_wire, list) and '__init__' in auto_wire:
                            auto_wire.append('build')
                            auto_wire.remove('__init__')
                    except KeyError:
                        pass

                wrapped_ = wrapper(wrapped, container=container, **inj_kwargs)

                if return_wrapped:
                    if wrapper in {register_build, register_external_build}:
                        try:
                            return container.get(wrapped_)
                        except DependencyInstantiationError as e:
                            raise TypeError() from e
                    else:
                        return wrapped_()

                if wrapper in {register, register_build, register_external_build}:
                    return container.get(wrapped_)
                elif wrapper == factory:
                    return container.get(MyService)
                elif wrapper == provider:
                    return container.providers[wrapped_]
                elif wrapper in {inject, wire_}:
                    return wrapped_()
                elif wrapper is resource:
                    return wrapped_()()
                else:
                    raise RuntimeError("Unsupported helper")

            if lazy:
                return test(container, create_instance=create)

            return test(container, instance=create())

        return f

    return decorator


@parametrize_injection(all_tests)
def test_basic_wiring(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is None


@parametrize_injection(metaclass_tests + class_tests, return_wrapped=True,
                       auto_wire=['__init__', 'method'])
def test_complex_wiring(container, instance: MyService):
    assert instance.method() is container.get(YetAnotherService)


@parametrize_injection(class_tests,
                       return_wrapped=True,
                       create_subclass=True,
                       auto_wire=['__init__', 'method'],
                       wire_super=True)
def test_subclass_wire_super(container, instance: MyService):
    assert instance.method() is container.get(YetAnotherService)


@parametrize_injection(class_tests,
                       lazy=True,
                       return_wrapped=True,
                       create_subclass=True,
                       auto_wire=['method'],
                       wire_super=False)
def test_subclass_no_wire_super(container, create_instance: Callable):
    with pytest.raises(TypeError):
        create_instance()


@parametrize_injection(all_tests, lazy=True, return_wrapped=True,
                       auto_wire=False)
def test_no_wiring(container, create_instance: Callable):
    with pytest.raises(TypeError):
        instance = create_instance()
        if callable(instance):
            instance()


@parametrize_injection(all_tests, lazy=True, return_wrapped=True,
                       use_type_hints=False)
def test_no_type_hints(container, create_instance: Callable):
    with pytest.raises(TypeError):
        instance = create_instance()
        if callable(instance):
            instance()


@parametrize_injection(all_tests, use_type_hints=['service'])
def test_type_hints_only_service(container, instance):
    assert instance.service is container.get(Service)
    assert instance.another_service is None


@parametrize_injection(all_tests, lazy=True, return_wrapped=True,
                       use_type_hints=['another_service'])
def test_type_hints_only_another_service(container, create_instance: Callable):
    with pytest.raises(TypeError):
        instance = create_instance()
        if callable(instance):
            instance()


@parametrize_injection(
    all_tests,
    dependencies=lambda s: AnotherService if s == 'service' else None
)
def test_dependencies_func_override(container, instance: MyService):
    assert instance.service is container.get(AnotherService)
    assert instance.another_service is None


@parametrize_injection(
    all_tests,
    dependencies=lambda s: AnotherService if s == 'another_service' else None
)
def test_dependencies_func(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is container.get(AnotherService)


@parametrize_injection(all_tests,
                       dependencies=dict(service=AnotherService))
def test_dependencies_dict_override(container, instance: MyService):
    assert instance.service is container.get(AnotherService)
    assert instance.another_service is None


@parametrize_injection(all_tests,
                       dependencies=dict(another_service=AnotherService))
def test_dependencies_dict(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is container.get(AnotherService)


@parametrize_injection(function_tests, dependencies=(AnotherService,))
def test_function_dependencies_tuple_override(container, instance: MyService):
    assert instance.service is container.get(AnotherService)
    assert instance.another_service is None


@parametrize_injection(function_tests, dependencies=(None, AnotherService))
def test_function_dependencies_tuple(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is container.get(AnotherService)


@parametrize_injection(all_tests, use_names=True)
def test_use_names_activated(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is container.get('another_service')


@parametrize_injection(all_tests, use_names=['another_service'])
def test_use_names_only_another_service(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is container.get('another_service')


@parametrize_injection(all_tests, use_names=['service'])
def test_use_names_only_service(container, instance: MyService):
    assert instance.service is container.get(Service)
    assert instance.another_service is None



from .core import inject
from .helpers import (factory, implements, LazyConstantsMeta, new_container, provider,
                      register, wire)
from .providers.lazy import LazyCall, LazyMethodCall
from .providers.factory import Build
from .providers.tag import Tag, Tagged, TaggedDependencies
from .utils import is_compiled


def __version__() -> str:  # pragma: no cover
    import pkg_resources
    try:
        return pkg_resources.get_distribution(__name__).version
    except pkg_resources.DistributionNotFound:  # pragma: no cover
        # package is not installed
        return ''


__all__ = ['Build',
           'factory',
           'implements',
           'inject',
           'is_compiled',
           'LazyCall',
           'LazyConstantsMeta',
           'LazyMethodCall',
           'new_container',
           'provider',
           'register',
           'Tag',
           'Tagged',
           'TaggedDependencies',
           'wire',
           'world']

world = new_container()



import pytest

from antidote.core import DependencyContainer
from antidote.helpers.constants import LazyConstantsMeta
from antidote.providers import LazyCallProvider, FactoryProvider


@pytest.fixture()
def container():
    c = DependencyContainer()
    c.register_provider(LazyCallProvider(container=c))
    c.register_provider(FactoryProvider(container=c))

    return c


def test_resource_meta(container: DependencyContainer):
    class Conf(metaclass=LazyConstantsMeta, container=container):
        A = 'a'
        B = 'b'

        def get(self, key):
            return key * 2

    assert 'aa' == container.get(Conf.A)
    assert 'bb' == container.get(Conf.B)

    conf = Conf()

    assert 'aa' == conf.A
    assert 'bb' == conf.B


def test_missing_get(container: DependencyContainer):
    with pytest.raises(ValueError):
        class Conf(metaclass=LazyConstantsMeta, container=container):
            A = 'a'


def test_private(container: DependencyContainer):
    class Conf(metaclass=LazyConstantsMeta, container=container):
        _A = 'a'

        b = 'b'

        def get(self, key):
            return key * 2

    conf = Conf()
    assert 'a' == conf._A
    assert 'b' == conf.b



import functools
from typing import Callable, Sequence

from .._internal.utils import SlotsReprMixin
from ..core import DependencyContainer
from ..exceptions import DependencyNotFoundError

compiled = False


class Injection(SlotsReprMixin):
    """
    Maps an argument name to its dependency and if the injection is required,
    which is equivalent to no default argument.
    """
    __slots__ = ('arg_name', 'required', 'dependency')

    def __init__(self, arg_name: str, required: bool, dependency):
        self.arg_name = arg_name
        self.required = required
        self.dependency = dependency


class InjectionBlueprint(SlotsReprMixin):
    """
    Stores all the injections for a function.
    """
    __slots__ = ('injections',)

    def __init__(self, injections: Sequence[Injection]):
        self.injections = injections


class InjectedWrapper:
    """
    Wrapper which injects all the dependencies not supplied in the passed
    arguments. An InjectionBlueprint is used to store the mapping of the
    arguments to their dependency if any and if the injection is required.
    """

    def __init__(self,
                 container: DependencyContainer,
                 blueprint: InjectionBlueprint,
                 wrapped: Callable,
                 skip_self: bool = False):
        """
        Args:
            container: current DependencyContainer
            blueprint: Injection blueprint for the underlying function
            wrapped:  real function to be called
            skip_self:  whether the first argument must be skipped. Used internally
        """
        self.__wrapped__ = wrapped
        self.__container = container
        self.__blueprint = blueprint
        self.__injection_offset = 1 if skip_self else 0
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args, **kwargs):
        kwargs = _inject_kwargs(
            self.__container,
            self.__blueprint,
            self.__injection_offset + len(args),
            kwargs
        )
        return self.__wrapped__(*args, **kwargs)

    def __get__(self, instance, owner):
        wrapped = self.__wrapped__.__get__(instance, owner)
        return functools.wraps(wrapped, updated=())(InjectedBoundWrapper(
            self.__container,
            self.__blueprint,
            wrapped,
            isinstance(self.__wrapped__, classmethod)
            or (not isinstance(self.__wrapped__, staticmethod) and instance is not None)
        ))

    def __getattr__(self, item):
        return getattr(self.__wrapped__, item)


class InjectedBoundWrapper(InjectedWrapper):
    """
    Behaves like Python bound methods. Unsure whether this is really necessary
    or not.
    """

    def __get__(self, instance, owner):
        return self  # pragma: no cover


def _inject_kwargs(container: DependencyContainer,
                   blueprint: InjectionBlueprint,
                   offset: int,
                   kwargs: dict) -> dict:
    """
    Does the actual injection of the dependencies. Used by InjectedCallableWrapper.
    """
    dirty_kwargs = False
    for injection in blueprint.injections[offset:]:
        if injection.dependency is not None and injection.arg_name not in kwargs:
            dependency_instance = container.provide(injection.dependency)
            if dependency_instance is not None:
                if not dirty_kwargs:
                    kwargs = kwargs.copy()
                    dirty_kwargs = True
                kwargs[injection.arg_name] = dependency_instance.instance
            elif injection.required:
                raise DependencyNotFoundError(injection.dependency)

    return kwargs



import pytest

from antidote.core import DependencyContainer
from antidote.providers.lazy import LazyCall, LazyCallProvider, LazyMethodCall
from antidote.providers.factory import FactoryProvider


@pytest.fixture
def container():
    return DependencyContainer()


@pytest.fixture
def service_provider(container):
    provider = FactoryProvider(container=container)
    container.register_provider(provider)
    return provider


@pytest.fixture
def lazy_provider(container):
    provider = LazyCallProvider(container=container)
    container.register_provider(provider)
    return provider


def test_lazy(lazy_provider: LazyCallProvider, service_provider: FactoryProvider):
    def func(x):
        return x

    obj = object()
    test = LazyCall(func)(obj)

    assert obj is lazy_provider.provide(test).instance


def test_lazy_singleton(lazy_provider: LazyCallProvider,
                        service_provider: FactoryProvider):
    def func(x):
        return x

    test = LazyCall(func, singleton=False)(1)
    assert False is lazy_provider.provide(test).singleton

    test2 = LazyCall(func, singleton=True)(1)
    assert True is lazy_provider.provide(test2).singleton


@pytest.mark.parametrize(
    'args,kwargs',
    [
        ((1,), dict(test=2)),
        ((), dict(x=32)),
        ((4,), {})
    ]
)
def test_args_kwargs(lazy_provider: LazyCallProvider,
                     service_provider: FactoryProvider,
                     args, kwargs):
    def func(*args_, **kwargs_):
        return args_, kwargs_

    test = LazyCall(func)(*args, **kwargs)

    assert (args, kwargs) == lazy_provider.provide(test).instance


def test_method_call(lazy_provider: LazyCallProvider,
                     service_provider: FactoryProvider):
    x = object()

    @service_provider.register_class
    class Test:
        def get(self, s):
            return id(s)

        A = LazyMethodCall(get)(x)
        B = LazyMethodCall('get')(x)

    assert id(x) == lazy_provider.provide(Test.A).instance
    assert id(x) == lazy_provider.provide(Test.B).instance


def test_method_same_instance(lazy_provider: LazyCallProvider,
                              service_provider: FactoryProvider):
    @service_provider.register_class
    class Test:
        def get(self):
            return self

        A = LazyMethodCall(get, singleton=True)
        B = LazyMethodCall(get, singleton=False)

    x = lazy_provider.provide(Test.A).instance
    assert x is lazy_provider.provide(Test.A).instance
    assert x is lazy_provider.provide(Test.B).instance


def test_method_singleton(lazy_provider: LazyCallProvider,
                          service_provider: FactoryProvider):
    @service_provider.register_class
    class Test:
        def get(self):
            return self

        A = LazyMethodCall(get, singleton=True)
        B = LazyMethodCall(get, singleton=False)

    assert Test.A is Test.A
    assert True is lazy_provider.provide(Test.A).singleton
    assert Test.B is not Test.B
    assert False is lazy_provider.provide(Test.B).singleton


def test_method_direct_call(lazy_provider: LazyCallProvider,
                            service_provider: FactoryProvider):
    @service_provider.register_class
    class Test:
        def get(self):
            return self

        A = LazyMethodCall(get)

    t = Test()
    assert t is t.A

    t2 = Test()
    assert t2 is t2.A


@pytest.mark.parametrize(
    'args,kwargs',
    [
        ((1,), dict(test=2)),
        ((), dict(x=32)),
        ((4,), {})
    ]
)
def test_method_args_kwargs(lazy_provider: LazyCallProvider,
                            service_provider: FactoryProvider,
                            args, kwargs):
    @service_provider.register_class
    class Test:
        def get(self, *args_, **kwargs_):
            return args_, kwargs_

        A = LazyMethodCall(get)(*args, **kwargs)

    assert (args, kwargs) == Test().A
    assert (args, kwargs) == lazy_provider.provide(Test.A).instance



# cython: language_level=3
# cython: boundscheck=False, wraparound=False
from antidote.core.container cimport DependencyContainer, DependencyInstance, DependencyProvider

cdef class Tag:
    cdef:
        readonly str name
        readonly dict _attrs

cdef class Tagged:
    cdef:
        readonly str name

cdef class TaggedDependencies:
    cdef:
        DependencyContainer _container
        object _lock
        list _dependencies
        list _tags
        list _instances

cdef class TagProvider(DependencyProvider):
    cdef:
        dict _dependency_to_tag_by_tag_name

    cpdef DependencyInstance provide(self, dependency)



import inspect
from enum import Enum
from typing import Callable, cast, TypeVar

from .._internal.default_container import get_default_container
from ..core import DependencyContainer
from ..providers import IndirectProvider

T = TypeVar('T', bound=type)


def implements(interface: type,
               *,
               state: Enum = None,
               container: DependencyContainer = None) -> Callable[[T], T]:
    """
    Class decorator declaring the underlying class as a (possible) implementation
    to be used by Antidote when requested the specified interface.

    For now, the underlying class needs to be decorated with @register.

    Args:
        interface: Interface implemented by the decorated class.
        state: If multiple implementations exist for an interface, an
            :py:class:`~enum.Enum` should be used to identify all the possible
            states the application may be. Each state should be associated with
            one implementation. At runtime Antidote will retrieve the state
            (the :py:class:`~enum.Enum`) class to determine the current state.
        container: :py:class:`~.core.container.DependencyContainer` from which
            the dependencies should be retrieved. Defaults to the global
            container if it is defined.

    Returns:
        The decorated class, unmodified.
    """
    container = container or get_default_container()

    def register_implementation(cls):
        if not inspect.isclass(cls):
            raise TypeError("implements must be applied on a class, "
                            "not a {}".format(type(cls)))

        if not issubclass(cls, interface):
            raise TypeError("{} does not implement {}.".format(cls, interface))

        interface_provider = cast(IndirectProvider,
                                  container.providers[IndirectProvider])
        interface_provider.register(interface, cls, state)

        return cls

    return register_implementation



import collections.abc as c_abc
from typing import Any, Iterable, Mapping, Set, Dict, Hashable

from .container import DependencyContainer, DependencyInstance
from .exceptions import DependencyNotFoundError


class ProxyContainer(DependencyContainer):
    """
    Proxy core which should only be used for mocking an testing.
    """

    def __init__(self,
                 container: DependencyContainer,
                 dependencies: Mapping = None,
                 include: Iterable = None,
                 exclude: Iterable = None,
                 missing: Iterable = None):
        super().__init__()
        for provider in container.providers.values():
            self.register_provider(provider)

        if missing is None:
            self._missing = set()  # type: Set[Any]
        elif isinstance(missing, c_abc.Iterable):
            self._missing = set(missing)
        else:
            raise ValueError("missing must be either an iterable or None")

        new_singletons = {}    # type: Dict[Any, DependencyInstance]
        if include is None:
            new_singletons = container.singletons
        elif isinstance(include, c_abc.Iterable):
            existing_singletons = container.singletons
            for dependency in include:
                new_singletons[dependency] = existing_singletons[dependency]
        else:
            raise ValueError("include must be either an iterable or None")

        if isinstance(exclude, c_abc.Iterable):
            for dependency in exclude:
                try:
                    del new_singletons[dependency]
                except KeyError:
                    pass
        elif exclude is not None:
            raise ValueError("exclude must be either an iterable or None")
        self.update_singletons({
            k: v.instance
            for k, v in new_singletons.items()
        })

        if isinstance(dependencies, c_abc.Mapping):
            self.update_singletons(dependencies)
        elif dependencies is not None:
            raise ValueError("dependencies must be either a mapping or None")

    def provide(self, dependency: Hashable):
        if dependency in self._missing:
            raise DependencyNotFoundError(dependency)

        return super().provide(dependency)



from itertools import chain


class SlotsReprMixin:
    """
    Used by all classes using __slots__ to generate an helpful repr with all
    the defined fields.
    """
    __slots__ = ()

    def __repr__(self):
        slots = chain.from_iterable(getattr(cls, '__slots__', [])
                                    for cls in type(self).__mro__)
        return "{type}({slots})".format(
            type=type(self).__name__,
            slots=', '.join((
                '{}={!r}'.format(name, getattr(self, name))
                for name in slots
            ))
        )



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from contextlib import contextmanager
from typing import Hashable

# @formatter:off
cimport cython
from cpython.set cimport PySet_Contains

from ..exceptions import DependencyCycleError
# @formatter:on

# Final class as its behavior can only be changed through a Cython class. The
# DependencyContainer calls directly push() and pop()
@cython.final
cdef class DependencyStack:
    def __init__(self):
        self._stack = list()
        self._seen = set()

    @contextmanager
    def instantiating(self, dependency: Hashable):
        if 1 != self.push(dependency):
            raise DependencyCycleError(self._stack.copy() + [dependency])
        try:
            yield
        finally:
            self.pop()

    cdef bint push(self, object dependency):
        """
        Args:
            dependency: supposed to be hashable as the core tries to 
                retrieves from a dictionary first. 

        Returns:
            0 if present in the stack
            1 OK
        """
        cdef:
            list stack
            bint seen

        if 1 == PySet_Contains(self._seen, dependency):
            return 0

        self._stack.append(dependency)
        self._seen.add(dependency)
        return 1

    cdef pop(self):
        """
        Latest elements of the stack is removed.
        """
        self._seen.remove(self._stack.pop())



import pytest

from antidote.core import DependencyContainer, ProxyContainer
from antidote.exceptions import DependencyNotFoundError
from .utils import DummyProvider


class Service:
    def __init__(self, *args):
        pass


class AnotherService:
    def __init__(self, *args):
        pass


def test_invalid_arguments():
    with pytest.raises(ValueError):
        ProxyContainer(DependencyContainer(), include=object())

    with pytest.raises(ValueError):
        ProxyContainer(DependencyContainer(), exclude=object())

    with pytest.raises(ValueError):
        ProxyContainer(DependencyContainer(), missing=object())

    with pytest.raises(ValueError):
        ProxyContainer(DependencyContainer(), dependencies=object())


def test_context_isolation():
    container = DependencyContainer()
    container.update_singletons({'test': 1})
    container.update_singletons({'name': 'Antidote'})
    s = Service()
    container.update_singletons({Service: s})

    proxy_container = ProxyContainer(container)

    assert s is proxy_container.get(Service)
    assert 1 == proxy_container.get('test')
    assert 'Antidote' == proxy_container.get('name')

    proxy_container.update_singletons({'another_service': AnotherService()})
    assert isinstance(proxy_container.get('another_service'), AnotherService)

    s2 = Service()
    proxy_container.update_singletons({Service: s2})
    assert s2 is proxy_container.get(Service)

    with pytest.raises(DependencyNotFoundError):
        container.get('another_service')

    assert s is container.get(Service)
    assert 1 == container.get('test')
    assert 'Antidote' == container.get('name')


def test_context_include():
    container = DependencyContainer()
    container.update_singletons({'test': 1})
    container.update_singletons({'name': 'Antidote'})
    s = Service()
    container.update_singletons({Service: s})

    proxy_container = ProxyContainer(container, include=[Service])

    assert s is proxy_container.get(Service)

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('name')

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('test')

    proxy_container = ProxyContainer(container, include=[])

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get(Service)

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('name')

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('test')


def test_context_exclude():
    container = DependencyContainer()
    container.update_singletons({'test': 1})
    container.update_singletons({'name': 'Antidote'})
    s = Service()
    container.update_singletons({Service: s})

    proxy_container = ProxyContainer(container, exclude=['name', 'unknown'])

    assert s is proxy_container.get(Service)
    assert 1 == proxy_container.get('test')

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('name')


def test_context_override():
    container = DependencyContainer()
    container.update_singletons({'test': 1})
    container.update_singletons({'name': 'Antidote'})
    s = Service()
    container.update_singletons({Service: s})

    proxy_container = ProxyContainer(container,
                                     dependencies=dict(test=2, name='testing'))

    assert s is proxy_container.get(Service)
    assert 2 == proxy_container.get('test')
    assert 'testing' == proxy_container.get('name')


def test_context_missing():
    container = DependencyContainer()
    container.update_singletons({'test': 1})
    container.register_provider(DummyProvider({'name': 'Antidote'}))
    s = Service()
    container.update_singletons({Service: s})

    proxy_container = ProxyContainer(container, missing=['name'])

    assert s is proxy_container.get(Service)
    assert 1 == proxy_container.get('test')

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('name')

    proxy_container = ProxyContainer(container, missing=['test'], include=[Service])

    assert s is proxy_container.get(Service)
    assert 'Antidote' == proxy_container.get('name')

    with pytest.raises(DependencyNotFoundError):
        proxy_container.get('test')



from enum import Enum

import pytest

from antidote import implements, register
from antidote.core import DependencyContainer
from antidote.providers import IndirectProvider, FactoryProvider


@pytest.fixture()
def container():
    c = DependencyContainer()
    c.register_provider(FactoryProvider(container=c))
    c.register_provider(IndirectProvider(container=c))

    return c


class IService:
    pass


class Profile(Enum):
    A = 'A'
    B = 'B'


def test_implements(container: DependencyContainer):
    @implements(IService, container=container)
    @register(container=container)
    class Service(IService):
        pass

    assert container.get(IService) is container.get(Service)


def test_implements_profile(container: DependencyContainer):
    @implements(IService, state=Profile.A, container=container)
    @register(container=container)
    class ServiceA(IService):
        pass

    @implements(IService, state=Profile.B, container=container)
    @register(container=container)
    class ServiceB(IService):
        pass

    container.update_singletons({Profile: Profile.A})
    assert container.get(IService) is container.get(ServiceA)


def test_invalid_implements(container: DependencyContainer):
    with pytest.raises(TypeError):
        @implements(IService, container=container)
        class A:
            pass

    with pytest.raises(TypeError):
        implements(IService, container=container)(1)



from .core.exceptions import (AntidoteError, DependencyCycleError,
                              DependencyInstantiationError, DependencyNotFoundError,
                              DuplicateDependencyError)


class DuplicateTagError(AntidoteError):
    """
    A dependency has multiple times the same tag.
    Raised by the :py:class:`~.providers.tag.TagProvider`.
    """


class UndefinedContextError(AntidoteError):
    """
    A context does not have any target associated.
    Raised by the :py:class:`~.providers.indirect.IndirectProvider`.
    """


__all__ = [
    'AntidoteError',
    'DependencyCycleError',
    'DependencyInstantiationError',
    'DependencyNotFoundError',
    'DuplicateDependencyError',
    'DuplicateTagError',
    'UndefinedContextError'
]



from antidote.exceptions import (DependencyCycleError, DependencyNotFoundError,
                                 DuplicateDependencyError)


class Service:
    pass


def test_dependency_cycle_error():
    error = DependencyCycleError([Service, 'test', 1, Service])

    service_info = "{}.{}".format(Service.__module__, Service.__name__)

    for f in [str, repr]:
        assert service_info in f(error)
        assert "'test'" in f(error)
        assert " 1 " in f(error)


def test_dependency_not_found():
    o = object()
    error = DependencyNotFoundError(o)

    for f in [str, repr]:
        assert repr(o) in f(error)


def test_duplicate_dependency_error():
    dependency = object()
    existing_dependency = object()
    error = DuplicateDependencyError(dependency, existing_dependency)

    for f in [str, repr]:
        assert f(dependency) in f(error)
        assert f(existing_dependency) in f(error)



import threading
from typing import (Any, cast, Dict, Generic, Hashable, List, Mapping, Optional, Tuple,
                    TypeVar)

from .exceptions import (DependencyCycleError, DependencyInstantiationError,
                         DependencyNotFoundError)
from .._internal.stack import DependencyStack
from .._internal.utils import SlotsReprMixin

T = TypeVar('T')


class DependencyInstance(SlotsReprMixin, Generic[T]):
    """
    Simple wrapper used by a :py:class:`~.core.Provider` when returning an
    instance of a dependency so it can specify in which scope the instance
    belongs to.
    """
    __slots__ = ('instance', 'singleton')

    def __init__(self, instance: T, singleton: bool = False):
        self.instance = instance
        self.singleton = singleton


class DependencyContainer:
    """
    Instantiates the dependencies through the registered providers and handles
    their scope.
    """

    def __init__(self):
        self._providers = list()  # type: List[DependencyProvider]
        self._type_to_provider = dict()  # type: Dict[type, DependencyProvider]
        self._singletons = dict()  # type: Dict[Any, DependencyInstance]
        self._singletons[DependencyContainer] = DependencyInstance(self, singleton=True)
        self._dependency_stack = DependencyStack()
        self._instantiation_lock = threading.RLock()

    def __str__(self):
        return "{}(providers=({}))".format(
            type(self).__name__,
            ", ".join("{}={}".format(name, p)
                      for name, p in self.providers.items()),
        )

    def __repr__(self):
        return "{}(providers=({}), singletons={!r})".format(
            type(self).__name__,
            ", ".join("{!r}={!r}".format(name, p)
                      for name, p in self.providers.items()),
            self._singletons
        )

    @property
    def providers(self) -> Mapping[type, 'DependencyProvider']:
        """ Returns a mapping of all the registered providers by their type. """
        return {type(p): p for p in self._providers}

    @property
    def singletons(self) -> dict:
        """ Returns all the defined singletons """
        return self._singletons.copy()

    def register_provider(self, provider: 'DependencyProvider'):
        """
        Registers a provider, which can then be used to instantiate dependencies.

        Args:
            provider: Provider instance to be registered.

        """
        if not isinstance(provider, DependencyProvider):
            raise TypeError("provider must be a DependencyProvider, not a {!r}".format(
                type(provider)
            ))

        for bound_type in provider.bound_dependency_types:
            if bound_type in self._type_to_provider:
                raise RuntimeError(
                    "Cannot bind {!r} to provider, already bound to {!r}".format(
                        bound_type, self._type_to_provider[bound_type]
                    ))

        for bound_type in provider.bound_dependency_types:
            self._type_to_provider[bound_type] = provider

        self._providers.append(provider)

    def update_singletons(self, dependencies: Mapping):
        """
        Update the singletons.
        """
        with self._instantiation_lock:
            self._singletons.update({
                k: DependencyInstance(v, singleton=True)
                for k, v in dependencies.items()
            })

    def get(self, dependency: Hashable):
        """
        Returns an instance for the given dependency. All registered providers
        are called sequentially until one returns an instance.  If none is
        found, :py:exc:`~.exceptions.DependencyNotFoundError` is raised.

        Args:
            dependency: Passed on to the registered providers.

        Returns:
            instance for the given dependency
        """
        return self.safe_provide(dependency).instance

    def safe_provide(self, dependency: Hashable) -> DependencyInstance:
        dependency_instance = self.provide(dependency)
        if dependency_instance is None:
            raise DependencyNotFoundError(dependency)
        return dependency_instance

    def provide(self, dependency: Hashable) -> Optional[DependencyInstance]:
        """
        Internal method which should not be directly called. Prefer
        :py:meth:`~.core.core.DependencyContainer.get`.
        It may be overridden in a subclass to customize how dependencies are
        instantiated.

        Used by the injection wrappers.
        """
        try:
            return self._singletons[dependency]
        except KeyError:
            pass

        try:
            # @formatter:off
            with self._instantiation_lock, \
                    self._dependency_stack.instantiating(dependency):
                # @formatter:on
                try:
                    return self._singletons[dependency]
                except KeyError:
                    pass

                dependency_instance = None
                provider = self._type_to_provider.get(type(dependency))
                if provider is not None:
                    dependency_instance = provider.provide(dependency)
                else:
                    for provider in self._providers:
                        dependency_instance = provider.provide(dependency)
                        if dependency_instance is not None:
                            break

                if dependency_instance is not None:
                    if dependency_instance.singleton:
                        self._singletons[dependency] = dependency_instance

                    return dependency_instance

        except DependencyCycleError:
            raise

        except Exception as e:
            raise DependencyInstantiationError(dependency) from e

        return None


class DependencyProvider:
    """
    Abstract base class for a Provider.

    Used by the :py:class:`~.core.DependencyContainer` to instantiate
    dependencies. Several are used in a cooperative manner : the first instance
    to be returned by one of them is used. Thus providers should ideally not
    overlap and handle only one kind of dependencies such as strings or tag.

    This should be used whenever one needs to introduce a new kind of dependency,
    or control how certain dependencies are instantiated.
    """
    bound_dependency_types = cast(Tuple[type], ())  # type: Tuple[type, ...]

    def __init__(self, container: DependencyContainer):
        self._container = container  # type: DependencyContainer

    def provide(self, dependency: Hashable) -> Optional[DependencyInstance]:
        """
        Method called by the :py:class:`~.core.DependencyContainer` when
        searching for a dependency.

        It is necessary to check quickly if the dependency can be provided or
        not, as :py:class:`~.core.DependencyContainer` will try its
        registered providers. A good practice is to subclass
        :py:class:`~.core.Dependency` so custom dependencies be differentiated.

        Args:
            dependency: The dependency to be provided by the provider.

        Returns:
            The requested instance wrapped in a :py:class:`~.core.Instance`
            if available or :py:obj:`None`.
        """
        raise NotImplementedError()  # pragma: no cover



from contextlib import contextmanager

from ..exceptions import DependencyCycleError


class DependencyStack:
    """
    Stores the stack of dependency instantiation to detect and prevent cycles
    by raising DependencyCycleError.

    Used in the DependencyContainer.

    This class is not thread-safe by itself.
    """

    def __init__(self):
        self._stack = list()
        self._seen = set()

    @contextmanager
    def instantiating(self, dependency):
        """
        Context Manager which has to be used when instantiating the
        dependency to keep track of the dependency path.

        When a cycle is detected, a DependencyCycleError is raised.
        """
        if dependency in self._seen:
            raise DependencyCycleError(self._stack + [dependency])

        self._stack.append(dependency)
        self._seen.add(dependency)
        try:
            yield
        finally:
            self._seen.remove(self._stack.pop())



from .container import new_container
from .factory import factory
from .provider import provider
from .register import register
from .constants import LazyConstantsMeta
from .wire import wire
from .implements import implements



import pytest

from antidote import factory
from antidote.core import DependencyContainer
from antidote.providers import FactoryProvider, TagProvider


@pytest.fixture()
def container():
    c = DependencyContainer()
    c.register_provider(FactoryProvider(container=c))
    c.register_provider(TagProvider(container=c))

    return c


class Service:
    pass


class AnotherService:
    pass


class YetAnotherService:
    pass


class SuperService:
    pass


def test_function(container):
    @factory(container=container)
    def build() -> Service:
        return Service()

    assert isinstance(container.get(Service), Service)
    # singleton by default
    assert container.get(Service) is container.get(Service)


def test_class(container):
    @factory(container=container)
    class ServiceFactory:
        def __call__(self) -> Service:
            return Service()

    assert isinstance(container.get(Service), Service)
    # singleton by default
    assert container.get(Service) is container.get(Service)


def test_missing_return_type_hint(container):
    with pytest.raises(ValueError):
        @factory(container=container)
        def faulty_service_provider():
            return Service()

    with pytest.raises(ValueError):
        @factory(container=container)
        class FaultyServiceFactory:
            def __call__(self):
                return Service()


@pytest.mark.parametrize('func', [1, type('MissingCall', tuple(), {})])
def test_invalid_func(func):
    with pytest.raises(TypeError):
        factory(func)



from ..core import DependencyContainer
from ..providers import LazyCallProvider, FactoryProvider, TagProvider, IndirectProvider


def new_container() -> DependencyContainer:
    container = DependencyContainer()
    container.register_provider(FactoryProvider(container))
    container.register_provider(LazyCallProvider(container))
    container.register_provider(TagProvider(container))
    container.register_provider(IndirectProvider(container))

    return container



from typing import Iterable, Union

from .register import register
from .wire import wire
from ..core import DEPENDENCIES_TYPE, DependencyContainer
from ..providers.lazy import LazyMethodCall


class LazyConstantsMeta(type):
    def __new__(metacls, cls, bases, namespace,
                lazy_method: str = 'get',
                auto_wire: Union[bool, Iterable[str]] = None,
                dependencies: DEPENDENCIES_TYPE = None,
                use_names: Union[bool, Iterable[str]] = None,
                use_type_hints: Union[bool, Iterable[str]] = None,
                container: DependencyContainer = None):
        """
        Metaclass used to generate class with constant dependencies.

        This should be used for configuration or external static resources.
        Only public uppercase class attributes will be converted to dependencies.

        .. doctest::

            >>> import antidote
            >>> class Conf(metaclass=antidote.LazyConstantsMeta):
            ...     DOMAIN = 'domain'
            ...     _A = 'unchanged'
            ...     a = 'unchanged'
            ...
            ...     def __init__(self):
            ...         self._data = {'domain': 'example.com'}
            ...
            ...     def get(self, key):
            ...         return self._data[key]
            ...
            >>> Conf._A
            'unchanged'
            >>> Conf.a
            'unchanged'
            >>> Conf().DOMAIN
            'example.com'
            >>> Conf.DOMAIN
            LazyMethodCallDependency(...)
            >>> antidote.world.get(Conf.DOMAIN)
            'example.com'
            >>> @antidote.inject(dependencies=(Conf.DOMAIN,))
            ... def f(a):
            ...     return a
            >>> f()
            'example.com'

        As one can see, neither :code:`a` nor :code:`_A` are changed,
        only :code:`DOMAIN`. Constant's initial value becomes the argument given
        to the lazy method, by default :code:`__call__()`. It has two different
        behaviors depending how it is retrieved:

        - Used as a instance attribute, :code:`Conf().DOMAIN`, is is equivalent
          to :code:`Conf().__call__('domain')`. This lets your code stay easy to
          manipulate and test.
        - Used as a class attribute, :code:`Conf.DOMAIN`, it becomes a special
          object used by Antidote to identify a dependency. This lets you inject
          :code:`Conf.DOMAIN` anywhere in your code.

        The advantage of using this is that Antidote will only instantiate
        :code:`Conf` once, if and only if necessary. The same is applied for
        every constant, those are singletons. Defining your static resources or
        configuration as class constants also makes your code more maintainable,
        as any decent IDE will refactor / find the usage of those in a blink of
        an eye.

        Underneath it uses :py:class:`.LazyMethodCall` and :py:func:`.register`.
        It is equivalent to:

        .. testcode::

            from antidote import LazyMethodCall, register

            @register(auto_wire=('__init__', '__call__'))
            class Conf:
                # Required for the example as we specify __init__() explicitly
                # for auto wiring, so it has to exist.
                def __init__(self):
                    pass

                def __call__(self, key):
                    return config[key]

                DOMAIN = LazyMethodCall(__call__)('domain')

        Args:
            lazy_method: Name of the lazy method to use for the constants.
                Defaults to :code:`'__call__'`.
            auto_wire: Injects automatically the dependencies of the methods
                specified, or only of :code:`__init__()` and :code:`__call__()`
                if True.
            dependencies: Can be either a mapping of arguments name to their
                dependency, an iterable of dependencies or a function which returns
                the dependency given the arguments name. If an iterable is specified,
                the position of the arguments is used to determine their respective
                dependency. An argument may be skipped by using :code:`None` as a
                placeholder. The first argument is always ignored for methods (self)
                and class methods (cls).Type hints are overridden. Defaults to
                :code:`None`.
            use_names: Whether or not the arguments' name should be used as their
                respective dependency. An iterable of argument names may also be
                supplied to restrict this to those. Defaults to :code:`False`.
            use_type_hints: Whether or not the type hints (annotations) should be
                used as the arguments dependency. An iterable of argument names may
                also be specified to restrict this to those. Any type hints from
                the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
                ...) are ignored. Defaults to :code:`True`.
            container: :py:class:`~.core.container.DependencyContainer` to which the
                dependency should be attached. Defaults to the global container,
                :code:`antidote.world`.
        """
        if lazy_method not in namespace:
            raise ValueError(
                "Lazy method {}() is no defined in {}".format(lazy_method, cls)
            )

        resource_class = super().__new__(metacls, cls, bases, namespace)

        wire_raise_on_missing = True
        if auto_wire is None or isinstance(auto_wire, bool):
            if auto_wire is False:
                methods = ()  # type: Iterable[str]
            else:
                methods = (lazy_method, '__init__')
                wire_raise_on_missing = False
        else:
            methods = auto_wire

        if methods:
            resource_class = wire(
                resource_class,
                methods=methods,
                dependencies=dependencies,
                use_names=use_names,
                use_type_hints=use_type_hints,
                container=container,
                raise_on_missing=wire_raise_on_missing
            )

        resource_class = register(
            resource_class,
            auto_wire=False,
            singleton=True,
            container=container
        )

        func = resource_class.__dict__[lazy_method]
        for name, v in list(resource_class.__dict__.items()):
            if not name.startswith('_') and name.isupper():
                setattr(resource_class, name, LazyMethodCall(func, singleton=True)(v))

        return resource_class

    # Python 3.5 compatibility
    def __init__(metacls, cls, bases, namespace, **kwargs):
        super().__init__(cls, bases, namespace)



# cython: language_level=3
# cython: boundscheck=False, wraparound=False
from antidote.core.container cimport DependencyInstance, DependencyProvider

cdef class IndirectProvider(DependencyProvider):
    cdef:
        dict _stateful_links
        dict _links

    cpdef DependencyInstance provide(self, object dependency)



import inspect
from typing import Callable, get_type_hints, Iterator, List, Sequence, Union


class Argument:
    def __init__(self, name: str, has_default: bool, type_hint):
        self.name = name
        self.has_default = has_default
        self.type_hint = type_hint

    def __repr__(self):
        common = "{}:{}".format(
            self.name,
            getattr(self.type_hint, "__name__", repr(self.type_hint))
        )
        return common + " = ?" if self.has_default else common


class Arguments:
    @classmethod
    def from_callable(cls,
                      func: Union[Callable, staticmethod, classmethod]
                      ) -> 'Arguments':
        unbound_method = is_unbound_method(func)  # doing it before un-wrapping.
        if isinstance(func, (staticmethod, classmethod)):
            func = func.__func__
        return cls._build(func, unbound_method)

    @classmethod
    def _build(cls, func: Callable, unbound_method: bool) -> 'Arguments':
        arguments = []  # type: List[Argument]
        has_var_positional = False
        has_var_keyword = False

        try:
            # typing is used, as lazy evaluation is not done properly with Signature.
            type_hints = get_type_hints(func)
        except Exception:  # Python 3.5.3 does not handle properly method wrappers
            type_hints = {}

        for name, parameter in inspect.signature(func).parameters.items():
            if parameter.kind is parameter.VAR_POSITIONAL:
                has_var_positional = True
            elif parameter.kind is parameter.VAR_KEYWORD:
                has_var_keyword = True
            else:
                arguments.append(Argument(
                    name=name,
                    has_default=parameter.default is not parameter.empty,
                    type_hint=type_hints.get(name)
                ))

        return Arguments(arguments=tuple(arguments),
                         has_var_positional=has_var_positional,
                         has_var_keyword=has_var_keyword,
                         has_self=unbound_method)

    def __init__(self,
                 arguments: Sequence[Argument],
                 has_var_positional: bool,
                 has_var_keyword: bool,
                 has_self: bool):
        self.arguments = arguments
        self.name_to_argument = dict(((arg.name, arg)
                                      for arg in arguments))
        self.has_var_positional = has_var_positional
        self.has_var_keyword = has_var_keyword
        self.has_self = has_self

        if has_self:
            self.without_self = Arguments(self.arguments[1:], self.has_var_positional,
                                          self.has_var_keyword, has_self=False)
        else:
            self.without_self = self

    def __repr__(self):
        args = [repr(arg) for arg in self.name_to_argument.values()]
        if self.has_var_positional:
            args.append("*args")
        if self.has_var_keyword:
            args.append("**kwargs")

        return "Arguments({})".format(", ".join(args))

    def __getitem__(self, index: Union[str, int]) -> Argument:
        if isinstance(index, str):
            return self.name_to_argument[index]
        elif isinstance(index, int):
            return self.arguments[index]
        else:
            raise TypeError("Unsupported index {!r}".format(index))

    def __contains__(self, name):
        return name in self.name_to_argument

    def __len__(self):
        return len(self.arguments)

    def __iter__(self) -> Iterator[Argument]:
        return iter(self.arguments)


def is_unbound_method(func: Union[Callable, staticmethod, classmethod]) -> bool:
    """
    Methods and nested function will have a different __qualname__ (See PEP-3155).

    >>> class A:
    ...     def f(self):
    ...         pass
    >>> A.f.__qualname__
    'A.f'

    With methods which still have self (or cls for class methods), we will ignore
    """
    if isinstance(func, staticmethod):
        return False

    if isinstance(func, classmethod):
        func = func.__func__

    return (func.__qualname__ != func.__name__  # not top level
            # not a bound method (self/cls already bound)
            and not inspect.ismethod(func)
            # not nested function
            and not func.__qualname__[:-len(func.__name__)].endswith("<locals>."))



from enum import Enum
from typing import Dict, Hashable, Optional

from .._internal.utils import SlotsReprMixin
from ..core import DependencyInstance, DependencyProvider
from ..exceptions import DuplicateDependencyError, UndefinedContextError


class IndirectProvider(DependencyProvider):
    """
    IndirectProvider
    """

    def __init__(self, container):
        super(IndirectProvider, self).__init__(container)
        self._stateful_links = dict()  # type: Dict[Hashable, StatefulLink]
        self._links = dict()  # type: Dict[Hashable, Hashable]

    def provide(self, dependency: Hashable) -> Optional[DependencyInstance]:
        try:
            target = self._links[dependency]
        except KeyError:
            try:
                stateful_link = self._stateful_links[dependency]
            except KeyError:
                return None
            else:
                state = self._container.safe_provide(
                    stateful_link.state_dependency
                )

                try:
                    target = stateful_link.targets[state.instance]
                except KeyError:
                    raise UndefinedContextError(dependency, state.instance)

                t = self._container.safe_provide(target)
                return DependencyInstance(
                    t.instance,
                    singleton=state.singleton & t.singleton
                )
        else:
            return self._container.safe_provide(target)

    def register(self, dependency: Hashable, target_dependency: Hashable,
                 state: Enum = None):
        if dependency in self._links:
            raise DuplicateDependencyError(dependency,
                                           self._links[dependency])

        if state is None:
            if dependency in self._stateful_links:
                raise DuplicateDependencyError(dependency,
                                               self._stateful_links[dependency])
            self._links[dependency] = target_dependency
        elif isinstance(state, Enum):
            try:
                stateful_link = self._stateful_links[dependency]
            except KeyError:
                stateful_link = StatefulLink(type(state))
                self._stateful_links[dependency] = stateful_link

            if state in stateful_link.targets:
                raise DuplicateDependencyError((dependency, state),
                                               stateful_link.targets[state])

            stateful_link.targets[state] = target_dependency
        else:
            raise TypeError("profile must be an instance of Flag or be None, "
                            "not a {!r}".format(type(state)))


class StatefulLink(SlotsReprMixin):
    """
    Internal API
    """
    __slots__ = ('state_dependency', 'targets')

    def __init__(self, state_dependency: Hashable):
        self.state_dependency = state_dependency
        self.targets = dict()  # type: Dict[Enum, Hashable]



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from enum import Enum
from typing import Hashable, Dict

# @formatter:off
from cpython.dict cimport PyDict_GetItem
from cpython.object cimport PyObject

from antidote.core.container cimport DependencyInstance, DependencyProvider
from ..exceptions import DuplicateDependencyError, UndefinedContextError
# @formatter:on


cdef class IndirectProvider(DependencyProvider):
    def __init__(self, container):
        super(IndirectProvider, self).__init__(container)
        self._stateful_links = dict()  # type: Dict[Hashable, StatefulLink]
        self._links = dict()  # type: Dict[Hashable, Hashable]

    cpdef DependencyInstance provide(self, object dependency):
        cdef:
            PyObject*ptr
            object service
            StatefulLink stateful_link
            DependencyInstance state
            DependencyInstance ContextualTarget_dependency

        ptr = PyDict_GetItem(self._links, dependency)
        if ptr != NULL:
            return self._container.safe_provide(<object> ptr)

        ptr = PyDict_GetItem(self._stateful_links, dependency)
        if ptr != NULL:
            stateful_link = <StatefulLink> ptr
            state = self._container.safe_provide(
                stateful_link.state_dependency
            )

            ptr = PyDict_GetItem(stateful_link.targets, state.instance)
            if ptr == NULL:
                raise UndefinedContextError(dependency, state.instance)

            target = self._container.safe_provide(<object> ptr)
            return DependencyInstance.__new__(
                DependencyInstance,
                target.instance,
                state.singleton & target.singleton
            )

        return None

    def register(self, dependency: Hashable, target_dependency: Hashable, state: Enum = None):
        cdef:
            StatefulLink stateful_link

        if dependency in self._links:
            raise DuplicateDependencyError(dependency,
                                           self._links[dependency])

        if state is None:
            if dependency in self._stateful_links:
                raise DuplicateDependencyError(dependency,
                                               self._stateful_links[dependency])
            self._links[dependency] = target_dependency
        elif isinstance(state, Enum):
            try:
                stateful_link = self._stateful_links[dependency]
            except KeyError:
                stateful_link = StatefulLink(type(state))
                self._stateful_links[dependency] = stateful_link

            if state in stateful_link.targets:
                raise DuplicateDependencyError((dependency, state),
                                               stateful_link.targets[state])

            stateful_link.targets[state] = target_dependency
        else:
            raise TypeError("profile must be an instance of Flag or be None, "
                            "not a {!r}".format(type(state)))

cdef class StatefulLink:
    cdef:
        object state_dependency
        dict targets

    def __init__(self, state_dependency):
        self.state_dependency = state_dependency
        self.targets = dict()  # type: Dict[Enum, Hashable]

    def __repr__(self):
        return "{}(state_dependency={!r}, targets={!r})".format(
            type(self).__name__,
            self.state_dependency,
            self.targets
        )



import typing

import pytest

from antidote._internal.argspec import Arguments
from antidote.core import DependencyContainer, inject
from antidote.exceptions import DependencyNotFoundError


class Service:
    pass


class AnotherService:
    pass


@pytest.mark.parametrize(
    'expected,kwargs',
    [
        pytest.param((None, None),
                     dict(),
                     id='nothing'),
        pytest.param((Service, None),
                     dict(dependencies=dict(first=Service)),
                     id='dependencies:dict-first'),
        pytest.param((Service, None),
                     dict(dependencies=(Service,)),
                     id='dependencies:tuple-first'),
        pytest.param((None, Service),
                     dict(dependencies=dict(second=Service)),
                     id='dependencies:dict-second'),
        pytest.param((None, Service),
                     dict(dependencies=(None, Service)),
                     id='dependencies:tuple-second'),
        pytest.param(('first', 'second'),
                     dict(dependencies=lambda s: s),
                     id='dependencies:callable'),
        pytest.param((Service, Service),
                     dict(dependencies=lambda s: Service),
                     id='dependencies:callable2'),
        pytest.param((None, None),
                     dict(dependencies=lambda s: None),
                     id='dependencies:callable3'),
        pytest.param(('first', 'second'),
                     dict(dependencies="{arg_name}"),
                     id='dependencies:str'),
        pytest.param(('prefix:first', 'prefix:second'),
                     dict(dependencies="prefix:{arg_name}"),
                     id='dependencies:str2'),
        pytest.param(('first', 'second'),
                     dict(use_names=True),
                     id='use_names:True'),
        pytest.param((None, None),
                     dict(use_names=False),
                     id='use_names:False'),
        pytest.param((None, 'second'),
                     dict(use_names=['second']),
                     id='use_names:list')
    ]
)
def test_without_type_hints(expected, kwargs):
    container = DependencyContainer()
    container.update_singletons({Service: Service()})
    container.update_singletons({AnotherService: AnotherService()})
    container.update_singletons({'first': object()})
    container.update_singletons({'second': object()})
    container.update_singletons({'prefix:first': object()})
    container.update_singletons({'prefix:second': object()})
    default = object()

    @inject(container=container, **kwargs)
    def f(first=default, second=default):
        return first, second

    class A:
        @inject(container=container, **kwargs)
        def method(self, first=default, second=default):
            return first, second

        @inject(container=container, **kwargs)
        @classmethod
        def class_method(cls, first=default, second=default):
            return first, second

        @inject(container=container, **kwargs)
        @staticmethod
        def static_method(first=default, second=default):
            return first, second

    expected = tuple((
        container.get(d) if d is not None else default
        for d in expected
    ))
    assert expected == f()
    assert expected == A().method()
    assert expected == A.class_method()
    assert expected == A.static_method()

    a, b = object(), object()
    assert (a, b) == f(a, b)
    assert (a, b) == A().method(a, b)
    assert (a, b) == A.class_method(a, b)
    assert (a, b) == A.static_method(a, b)


@pytest.mark.parametrize(
    'expected, kwargs',
    [
        pytest.param((Service, None),
                     dict(),
                     id='nothing'),
        pytest.param((Service, None),
                     dict(dependencies=dict(first=Service)),
                     id='dependencies:dict-first'),
        pytest.param((Service, None),
                     dict(dependencies=(Service,)),
                     id='dependencies:tuple-first'),
        pytest.param((Service, Service),
                     dict(dependencies=dict(second=Service)),
                     id='dependencies:dict-second'),
        pytest.param((Service, Service),
                     dict(dependencies=(None, Service)),
                     id='dependencies:tuple-second'),
        pytest.param(('first', 'second'),
                     dict(dependencies=lambda s: s),
                     id='dependencies:callable'),
        pytest.param((Service, Service),
                     dict(dependencies=lambda s: Service),
                     id='dependencies:callable2'),
        pytest.param((Service, None),
                     dict(dependencies=lambda s: None),
                     id='dependencies:callable3'),
        pytest.param(('first', 'second'),
                     dict(dependencies="{arg_name}"),
                     id='dependencies:str'),
        pytest.param(('prefix:first', 'prefix:second'),
                     dict(dependencies="prefix:{arg_name}"),
                     id='dependencies:str2'),
        pytest.param((Service, 'second'),
                     dict(use_names=True),
                     id='use_names:True'),
        pytest.param((Service, None),
                     dict(use_names=False),
                     id='use_names:False'),
        pytest.param((Service, None),
                     dict(use_names=['first']),
                     id='use_names:list-first'),
        pytest.param((Service, 'second'),
                     dict(use_names=['second']),
                     id='use_names:list-second'),
        pytest.param((Service, None),
                     dict(use_type_hints=True),
                     id='use_type_hints:True'),
        pytest.param((Service, None),
                     dict(use_type_hints=['first']),
                     id='use_type_hints:list-first'),
        pytest.param((Service, 'second'),
                     dict(use_type_hints=['first'], use_names=True),
                     id='use_type_hints:list-first+use_names=True'),
        pytest.param((None, None),
                     dict(use_type_hints=['second']),
                     id='use_type_hints:list-second'),
        pytest.param(('first', 'second'),
                     dict(use_type_hints=['second'], use_names=True),
                     id='use_type_hints:list-second+use_names=True'),
        pytest.param((None, None),
                     dict(use_type_hints=False),
                     id='use_type_hints:False'),
        pytest.param(('first', 'second'),
                     dict(use_type_hints=False, use_names=True),
                     id='use_type_hints:False+use_names=True'),
    ]
)
def test_with_type_hints(expected, kwargs):
    container = DependencyContainer()
    container.update_singletons({Service: Service(),
                                 AnotherService: AnotherService(),
                                 'first': object(),
                                 'second': object(),
                                 'prefix:first': object(),
                                 'prefix:second': object()})
    default = object()

    @inject(container=container, **kwargs)
    def f(first: Service = default, second: str = default):
        return first, second

    class A:
        @inject(container=container, **kwargs)
        def method(self, first: Service = default, second: str = default):
            return first, second

        @inject(container=container, **kwargs)
        @classmethod
        def class_method(cls, first: Service = default, second: str = default):
            return first, second

        @inject(container=container, **kwargs)
        @staticmethod
        def static_method(first: Service = default, second: str = default):
            return first, second

    expected = tuple((
        container.get(d) if d is not None else default
        for d in expected
    ))
    assert expected == f()
    assert expected == A().method()
    assert expected == A.class_method()
    assert expected == A.static_method()

    a, b = object(), object()
    assert (a, b) == f(a, b)
    assert (a, b) == A().method(a, b)
    assert (a, b) == A.class_method(a, b)
    assert (a, b) == A.static_method(a, b)


@pytest.mark.parametrize(
    'type_hint',
    [str, int, float, set, list, dict, complex, type, tuple, bytes, bytearray,
     typing.Optional, typing.Sequence]
)
def test_ignored_type_hints(type_hint):
    container = DependencyContainer()
    container.update_singletons({type_hint: object()})

    @inject(container=container)
    def f(x: type_hint):
        pass

    with pytest.raises(TypeError):
        f()


def test_arguments():
    container = DependencyContainer()
    container.update_singletons(dict(a=12, b=24))

    def f(a, b):
        pass

    arguments = Arguments.from_callable(f)

    @inject(arguments=arguments, use_names=True, container=container)
    def g(**kwargs):
        return kwargs

    assert dict(a=12, b=24) == g()


@pytest.mark.parametrize(
    'error,kwargs',
    [
        pytest.param(TypeError,
                     dict(),
                     id="unknown-dependency"),
        pytest.param(DependencyNotFoundError,
                     dict(dependencies=(Service,)),
                     id="dependencies:unknown-dependency-tuple"),
        pytest.param(DependencyNotFoundError,
                     dict(dependencies=dict(x=Service)),
                     id="dependencies:unknown-dependency-dict"),
        pytest.param(DependencyNotFoundError,
                     dict(dependencies=lambda s: Service),
                     id="dependencies:unknown-dependency-callable"),
        pytest.param(DependencyNotFoundError,
                     dict(dependencies="unknown:{arg_name}"),
                     id="dependencies:unknown-dependency-str"),
        pytest.param((ValueError, TypeError),
                     dict(dependencies=(None, None)),
                     id="dependencies:too-much-arguments"),
        pytest.param(TypeError,
                     dict(dependencies=object()),
                     id="dependencies:unsupported-type"),
        pytest.param(TypeError,
                     dict(dependencies={1: 'x'}),
                     id="dependencies:invalid-key-type"),
        pytest.param(ValueError,
                     dict(dependencies=dict(unknown=DependencyContainer)),
                     id="dependencies:unknown-argument-dict"),
        pytest.param(TypeError,
                     dict(use_names=False),
                     id="use_names:unknown-dependency-False"),
        pytest.param(DependencyNotFoundError,
                     dict(use_names=True),
                     id="use_names:unknown-dependency-True"),
        pytest.param(DependencyNotFoundError,
                     dict(use_names=['x']),
                     id="use_names:unknown-dependency-list"),
        pytest.param(ValueError,
                     dict(use_names=['y']),
                     id="use_names:unknown-argument-list"),
        pytest.param(ValueError,
                     dict(use_names=['x', 'y']),
                     id="use_names:unknown-argument-list2"),
        pytest.param(TypeError,
                     dict(use_names=[]),
                     id="use_names:empty"),
        pytest.param(TypeError,
                     dict(use_names=object()),
                     id="use_names:unsupported-type"),
        pytest.param(TypeError,
                     dict(use_names=[1]),
                     id="use_names:invalid-name-type"),
        pytest.param(TypeError,
                     dict(use_type_hints=object()),
                     id="use_type_hints:unsupported-type"),
        pytest.param(TypeError,
                     dict(use_type_hints=[1]),
                     id="use_type_hints:invalid-name-type"),
        pytest.param(ValueError,
                     dict(use_type_hints=['y']),
                     id="use_type_hints:unknown-arg"),
    ]
)
def test_invalid(error, kwargs):
    container = DependencyContainer()

    with pytest.raises(error):
        @inject(container=container, **kwargs)
        def f(x):
            return x

        f()

    with pytest.raises(error):
        class A:
            @inject(container=container, **kwargs)
            def method(self, x):
                return x

        A().method()

    with pytest.raises(error):
        class A:
            @inject(container=container, **kwargs)
            @classmethod
            def classmethod(cls, x):
                return x

        A.classmethod()

    with pytest.raises(error):
        class A:
            @inject(container=container, **kwargs)
            @staticmethod
            def staticmethod(x):
                return x

        A.staticmethod()


@pytest.mark.parametrize(
    'error,kwargs',
    [
        pytest.param(ValueError,
                     dict(dependencies=dict(self='x')),
                     id="dependencies"),
        pytest.param(ValueError,
                     dict(use_names=('self',)),
                     id="use_names"),
        pytest.param(ValueError,
                     dict(use_type_hints=('self',)),
                     id="use_type_hints"),
    ]
)
def test_cannot_inject_self(error, kwargs):
    container = DependencyContainer()
    container.update_singletons(dict(x=object(), y=object()))

    with pytest.raises(error):
        class A:
            @inject(container=container, **kwargs)
            def method(self, x=None):
                return x
        A()

    with pytest.raises(error):
        class A:
            @inject(container=container, **kwargs)
            @classmethod
            def classmethod(self, x=None):
                return x
        A()


def test_invalid_type_hint():
    @inject(container=DependencyContainer())
    def f(x: Service):
        return x

    with pytest.raises(DependencyNotFoundError):
        f()


def test_no_injections():
    container = DependencyContainer()

    def f(x):
        return x

    injected_f = inject(f, container=container)

    # When nothing can be injected, the same function should be returned
    assert injected_f is f


def test_already_injected():
    container = DependencyContainer()

    @inject(container=container, use_names=True)
    def f(x):
        return x

    injected_f = inject(f, container=container)

    # When the function has already its arguments injected, the same function should
    # be returned
    assert injected_f is f


def test_class_inject():
    container = DependencyContainer()
    with pytest.raises(TypeError):
        @inject(container=container)
        class Dummy:
            pass



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False

# @formatter:off
cimport cython
from cpython.dict cimport PyDict_Contains, PyDict_Copy, PyDict_SetItem
from cpython.object cimport PyObject_Call
from cpython.tuple cimport PyTuple_GET_ITEM,  PyTuple_Size

from antidote.core.container cimport DependencyContainer, DependencyInstance
from ..exceptions import DependencyNotFoundError
# @formatter:on

compiled = True

@cython.freelist(128)
cdef class Injection:
    cdef:
        readonly str arg_name
        readonly bint required
        readonly object dependency

    def __repr__(self):
        return "{}(arg_name={!r}, required={!r}, dependency={!r})".format(
            type(self).__name__,
            self.arg_name,
            self.required,
            self.dependency
        )

    def __init__(self, str arg_name, bint required, object dependency):
        self.arg_name = arg_name
        self.required = required
        self.dependency = dependency

cdef class InjectionBlueprint:
    cdef:
        readonly tuple injections

    def __init__(self, tuple injections):
        self.injections = injections

cdef class InjectedWrapper:
    cdef:
        readonly object __wrapped__
        DependencyContainer __container
        InjectionBlueprint __blueprint
        int __injection_offset
        dict __dict

    def __cinit__(self,
                  DependencyContainer container,
                  InjectionBlueprint blueprint,
                  object wrapped,
                  bint skip_first = False):
        self.__wrapped__ = wrapped
        self.__container = container
        self.__blueprint = blueprint
        self.__injection_offset = 1 if skip_first else 0
        # allocate a dictionary only if necessary, it's not that common
        # to add attributes to a function.
        self.__dict = None

    def __call__(self, *args, **kwargs):
        kwargs = _inject_kwargs(
            self.__container,
            self.__blueprint,
            self.__injection_offset + len(args),
            kwargs
        )
        return PyObject_Call(self.__wrapped__, args, kwargs)

    def __get__(self, instance, owner):
        return InjectedBoundWrapper.__new__(
            InjectedBoundWrapper,
            self.__container,
            self.__blueprint,
            self.__wrapped__.__get__(instance, owner),
            isinstance(self.__wrapped__, classmethod)
            or (not isinstance(self.__wrapped__, staticmethod) and instance is not None)
        )

    def __getattr__(self, name):
        if self.__dict is not None:
            try:
                return self.__dict[name]
            except KeyError:
                pass
        return getattr(self.__wrapped__, name)

    def __setattr__(self, name, value):
        if self.__dict is None:
            self.__dict = dict()
        self.__dict[name] = value

    def __delattr__(self, name):
        if self.__dict is not None:
            try:
                del self.__dict[name]
            except KeyError:
                raise AttributeError(name)

cdef class InjectedBoundWrapper(InjectedWrapper):
    def __get__(self, instance, owner):
        return self

cdef inline dict _inject_kwargs(DependencyContainer container,
                                InjectionBlueprint blueprint,
                                int offset,
                                dict kwargs):
    cdef:
        Injection injection
        DependencyInstance dependency_instance
        bint dirty_kwargs = False
        int i

    for i in range(offset, PyTuple_Size(blueprint.injections)):
        injection = <Injection> PyTuple_GET_ITEM(blueprint.injections, i)
        if injection.dependency is not None \
                and PyDict_Contains(kwargs, injection.arg_name) == 0:
            dependency_instance = container.provide(injection.dependency)
            if dependency_instance is not None:
                if not dirty_kwargs:
                    kwargs = PyDict_Copy(kwargs)
                    dirty_kwargs = True
                PyDict_SetItem(kwargs, injection.arg_name, dependency_instance.instance)
            elif injection.required:
                raise DependencyNotFoundError(injection.dependency)

    return kwargs



"""
Test only that the wrapper behaves nicely in all cases.
Injection itself is tested through inject.
"""
import inspect
from typing import Any, List, Tuple

import pytest

from antidote._internal.wrapper import InjectedWrapper, Injection, InjectionBlueprint
from antidote.core import DependencyContainer
from antidote.exceptions import DependencyNotFoundError

A = object()
B = object()
C = object()
default_container = DependencyContainer()
default_container.update_singletons(dict(x=A))


def easy_wrap(func=None,
              arg_dependency: List[Tuple[str, bool, Any]] = tuple(),
              container: DependencyContainer = None):
    def wrapper(func):
        return InjectedWrapper(
            container=container or default_container,
            blueprint=InjectionBlueprint(tuple([
                Injection(arg_name, required, dependency)
                for arg_name, required, dependency in arg_dependency
            ])),
            wrapped=func
        )

    return func and wrapper(func) or wrapper


arg_self_x = [('self', True, None), ('x', True, 'x')]
arg_cls_x = [('cls', True, None), ('x', True, 'x')]
arg_x = [('x', True, 'x')]


class Dummy:
    @easy_wrap(arg_dependency=arg_self_x)
    def method(self, x):
        return self, x

    @easy_wrap(arg_dependency=arg_cls_x)
    @classmethod
    def class_before(cls, x):
        return cls, x

    @classmethod
    @easy_wrap(arg_dependency=arg_cls_x)
    def class_after(cls, x):
        return cls, x

    @easy_wrap(arg_dependency=arg_x)
    @staticmethod
    def static_before(x):
        return x

    @staticmethod
    @easy_wrap(arg_dependency=arg_x)
    def static_after(x):
        return x


class Dummy2:
    def method(self, x):
        return self, x

    @classmethod
    def class_method(cls, x):
        return cls, x

    @staticmethod
    def static(x):
        return x


Dummy2.method = easy_wrap(Dummy2.__dict__['method'], arg_self_x)
Dummy2.class_method = easy_wrap(Dummy2.__dict__['class_method'], arg_cls_x)
Dummy2.static = easy_wrap(Dummy2.__dict__['static'], arg_x)


@easy_wrap(arg_dependency=arg_x)
def f(x):
    return x


d = Dummy()
d2 = Dummy2()


@pytest.mark.parametrize(
    'expected, func',
    [
        pytest.param(A, f,
                     id='func'),

        pytest.param((B, A), Dummy.method,
                     id='method'),
        pytest.param((Dummy, A), Dummy.class_before,
                     id='classmethod before'),
        pytest.param((Dummy, A), Dummy.class_after,
                     id='classmethod after'),
        pytest.param(A, Dummy.static_before,
                     id='staticmethod before'),
        pytest.param(A, Dummy.static_after,
                     id='staticmethod after'),

        pytest.param((d, A), d.method,
                     id='instance method'),
        pytest.param((Dummy, A), d.class_before,
                     id='instance classmethod before'),
        pytest.param((Dummy, A), d.class_after,
                     id='instance classmethod after'),
        pytest.param(A, d.static_before,
                     id='instance staticmethod before'),
        pytest.param(A, d.static_after,
                     id='instance staticmethod after'),

        pytest.param((d2, A), d2.method,
                     id='post:instance method'),
        pytest.param((Dummy2, A), d2.class_method,
                     id='post:instance classmethod'),
        pytest.param(A, d2.static,
                     id='post:instance staticmethod'),

        pytest.param((B, A), Dummy2.method,
                     id='post:method'),
        pytest.param((Dummy2, A), Dummy2.class_method,
                     id='post:classmethod'),
        pytest.param(A, Dummy2.static,
                     id='post:staticmethod'),
    ]
)
def test_wrapper(expected, func):
    if expected == (B, A):
        assert expected == func(B, A)
        assert expected == func(B)
        assert (B, C) == func(B, C)
        assert (B, C) == func(B, x=C)
    else:
        assert expected == func(A)
        assert expected == func()

        if isinstance(expected, tuple):
            new_expected = (expected[0], C)
        else:
            new_expected = C

        assert new_expected == func(C)
        assert new_expected == func(x=C)


def test_classmethod_wrapping():
    def class_method(cls):
        pass

    class A:
        method = easy_wrap(classmethod(class_method))

    assert class_method == A.__dict__['method'].__func__
    assert A == A.method.__self__


def test_required_dependency_not_found():
    @easy_wrap(arg_dependency=[('x', True, 'unknown')])
    def f(x):
        return x

    with pytest.raises(DependencyNotFoundError):
        f()


def test_dependency_not_found():
    @easy_wrap(arg_dependency=[('x', False, 'unknown')])
    def f(x):
        return x

    with pytest.raises(TypeError):
        f()


def test_multiple_injections():
    container = DependencyContainer()
    xx = object()
    yy = object()
    zz = object()
    container.update_singletons(dict(xx=xx, yy=yy))

    @easy_wrap(arg_dependency=[('x', True, 'xx'),
                               ('y', True, 'yy'),
                               ('z', False, 'zz')],
               container=container)
    def f(x, y, z=zz):
        return x, y, z

    assert (xx, yy, zz) == f()
    assert (xx, A, zz) == f(y=A)
    assert (xx, yy, A) == f(z=A)
    assert (A, yy, zz) == f(x=A)
    assert (A, yy, B) == f(A, z=B)

    with pytest.raises(TypeError):
        f(A, x=A)


def g():
    pass


def g_with_annotations(x: int):
    pass


def g_with_defaults(x, y=3):
    pass


def g_with_kw_only(x, *, y):
    pass


class G:
    def method(self):
        pass

    def method_with_annotations(self, x: int):
        pass

    def method_with_defaults(self, x, y=3):
        pass

    def method_with_kw_only(self, x, *, y):
        pass

    @staticmethod
    def cls():
        pass

    @staticmethod
    def cls_with_annotations(cls, x: int):
        pass

    @staticmethod
    def cls_with_defaults(cls, x, y=3):
        pass

    @staticmethod
    def cls_with_kw_only(cls, x, *, y):
        pass

    @staticmethod
    def static():
        pass

    @staticmethod
    def static_with_annotations(x: int):
        pass

    @staticmethod
    def static_with_defaults(x, y=3):
        pass

    @staticmethod
    def static_with_kw_only(x, *, y):
        pass


@pytest.mark.parametrize(
    'original,wrapped',
    [
        pytest.param(g,
                     easy_wrap(g),
                     id='function'),
        pytest.param(g_with_annotations,
                     easy_wrap(g_with_annotations,
                               arg_dependency=[('x', True, None)]),
                     id='function:__annotations__'),
        pytest.param(g_with_defaults,
                     easy_wrap(g_with_defaults,
                               arg_dependency=[('x', True, None), ('y', False, None)]),
                     id='function:__defaults__'),
        pytest.param(g_with_kw_only,
                     easy_wrap(g_with_kw_only,
                               arg_dependency=[('x', True, None), ('y', True, None)]),
                     id='function:__kwdefaults__'),
        pytest.param(G.method,
                     easy_wrap(G.method),
                     id='method'),
        pytest.param(G.method_with_annotations,
                     easy_wrap(G.method_with_annotations,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None)]),
                     id='method:__annotations__'),
        pytest.param(G.method_with_defaults,
                     easy_wrap(G.method_with_defaults,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None),
                                               ('y', False, None)]),
                     id='method:__defaults__'),
        pytest.param(G.method_with_kw_only,
                     easy_wrap(G.method_with_kw_only,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None),
                                               ('y', True, None)]),
                     id='method:__kwdefaults__'),
        pytest.param(G.cls,
                     easy_wrap(G.cls),
                     id='classmethod'),
        pytest.param(G.cls_with_annotations,
                     easy_wrap(G.cls_with_annotations,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None)]),
                     id='classmethod:__annotations__'),
        pytest.param(G.cls_with_defaults,
                     easy_wrap(G.cls_with_defaults,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None),
                                               ('y', False, None)]),
                     id='classmethod:__defaults__'),
        pytest.param(G.cls_with_kw_only,
                     easy_wrap(G.cls_with_kw_only,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None),
                                               ('y', True, None)]),
                     id='classmethod:__kwdefaults__'),
        pytest.param(G.static,
                     easy_wrap(G.static),
                     id='static'),
        pytest.param(G.static_with_annotations,
                     easy_wrap(G.static_with_annotations,
                               arg_dependency=[('x', True, None)]),
                     id='static:__annotations__'),
        pytest.param(G.static_with_defaults,
                     easy_wrap(G.static_with_defaults,
                               arg_dependency=[('x', True, None), ('y', False, None)]),
                     id='static:__defaults__'),
        pytest.param(G.static_with_kw_only,
                     easy_wrap(G.static_with_kw_only,
                               arg_dependency=[('x', True, None), ('y', True, None)]),
                     id='static:__kwdefaults__')
    ]
)
def test_wrapped_attributes(original, wrapped):
    assert original is wrapped.__wrapped__
    assert original.__module__ == wrapped.__module__
    assert original.__name__ == wrapped.__name__
    assert original.__qualname__ == wrapped.__qualname__
    assert original.__doc__ == wrapped.__doc__
    assert original.__annotations__ == wrapped.__annotations__
    assert original.__defaults__ == wrapped.__defaults__
    assert original.__kwdefaults__ == wrapped.__kwdefaults__

    try:
        func = original.__func__
    except AttributeError:
        with pytest.raises(AttributeError):
            wrapped.__func__
    else:
        assert func is wrapped.__func__

    try:
        func = original.__self__
    except AttributeError:
        with pytest.raises(AttributeError):
            wrapped.__self__
    else:
        assert func is wrapped.__self__

    assert inspect.signature(original) == inspect.signature(wrapped)


def test_custom_attributes():
    def f():
        pass

    f.attr = "test"
    f.another_attr = "another_attr"
    wrapped = easy_wrap(f)
    assert "test" == wrapped.attr

    wrapped.attr2 = "test2"
    assert "test2" == wrapped.attr2
    # After setting a new attribute, original ones should still be accessible.
    assert "test" == wrapped.attr

    wrapped.attr = "overridden test"
    assert "overridden test" == wrapped.attr

    # You should be able to remove new attributes
    del wrapped.attr2

    with pytest.raises(AttributeError):
        wrapped.attr2

    # but not existing ones
    with pytest.raises(AttributeError):
        del wrapped.another_attr



import pytest

from antidote import wire
from antidote.core import DependencyContainer


@pytest.fixture()
def container():
    container = DependencyContainer()
    container.update_singletons(dict(x=object(), y=object()))
    return container


def test_multi_wire(container: DependencyContainer):
    xx = container.get('x')
    yy = container.get('y')

    @wire(methods=['f', 'g'],
          dependencies=('x', 'y'),
          container=container)
    class Dummy:
        def f(self, x):
            return x

        def g(self, x, y):
            return x, y

    d1 = Dummy()
    assert xx == d1.f()
    assert (xx, yy) == d1.g()

    @wire(methods=['f', 'g'],
          dependencies=dict(x='x', y='y'),
          container=container)
    class Dummy:
        def f(self, x):
            return x

        def g(self, x, y):
            return x, y

    d1 = Dummy()
    assert xx == d1.f()
    assert (xx, yy) == d1.g()

    @wire(methods=['f', 'g'],
          use_names=['x', 'y'],
          container=container)
    class Dummy2:
        def f(self, x):
            return x

        def g(self, x, y):
            return x, y

    d2 = Dummy2()
    assert xx == d2.f()
    assert (xx, yy) == d2.g()

    container.update_singletons({Dummy: d1, Dummy2: d2})

    @wire(methods=['f', 'g'],
          use_type_hints=['x', 'y'],
          container=container)
    class Dummy3:
        def f(self, x: Dummy):
            return x

        def g(self, x: Dummy, y: Dummy2):
            return x, y

    assert d1 == Dummy3().f()
    assert (d1, d2) == Dummy3().g()


def test_subclass_classmethod(container: DependencyContainer):
    xx = container.get('x')

    @wire(methods=['cls_method'], use_names=True, container=container)
    class Dummy:
        @classmethod
        def cls_method(cls, x):
            return cls, x

    assert (Dummy, xx) == Dummy.cls_method()

    class SubDummy(Dummy):
        pass

    assert (SubDummy, xx) == SubDummy.cls_method()


def test_wire_super(container: DependencyContainer):
    xx = container.get('x')
    sentinel = object()

    class Dummy:
        def method(self, x):
            return self, x

    @wire(wire_super=True, methods=['method'], use_names=True, container=container)
    class SubDummy(Dummy):
        pass

    sub_dummy = SubDummy()

    assert (sub_dummy, xx) == sub_dummy.method()
    assert (sub_dummy, sentinel) == sub_dummy.method(sentinel)

    with pytest.raises(TypeError):  # did not affect base class
        Dummy().method()


def test_do_not_change_for_nothing(container: DependencyContainer):
    def original_method(self, something):
        pass

    @wire(methods=['method'], container=container)
    class Dummy:
        method = original_method

    assert Dummy.__dict__['method'] is original_method

    @wire(wire_super=True, methods=['method'], container=container)
    class SubDummy(Dummy):
        pass

    assert 'method' not in SubDummy.__dict__
    assert SubDummy.method is Dummy.method


@pytest.mark.parametrize('obj', [object(), lambda: None])
def test_invalid_class(obj):
    with pytest.raises(TypeError):
        wire(obj, methods=['__init__'])


@pytest.mark.parametrize(
    'kwargs',
    [
        dict(methods=['__init__'], wire_super=['__call__']),
    ]
)
def test_invalid_params(kwargs):
    with pytest.raises(ValueError):
        @wire(**kwargs)
        class Dummy:
            pass


@pytest.mark.parametrize(
    'kwargs',
    [
        dict(methods=object()),
        dict(methods=['method'], wire_super=object()),
        dict(methods=['method'], raise_on_missing=object()),
    ]
)
def test_invalid_type(kwargs):
    with pytest.raises(TypeError):
        @wire(**kwargs)
        class Dummy:
            pass


def test_raise_on_missing(container: DependencyContainer):
    with pytest.raises(TypeError):
        @wire(methods=['method'], container=container,
              raise_on_missing=True)
        class Dummy:
            pass

    @wire(methods=['method'], container=container, raise_on_missing=False)
    class Dummy2:
        pass

    @wire(wire_super=True, methods=['method'], container=container,
          raise_on_missing=False)
    class Dummy3(Dummy2):
        pass



import collections.abc as c_abc
import inspect
from typing import Callable, Iterable, Optional, overload, Set, TypeVar, Union

from .._internal.argspec import Arguments
from ..core import DEPENDENCIES_TYPE, DependencyContainer, inject

C = TypeVar('C', bound=type)


@overload
def wire(class_: C,  # noqa: E704  # pragma: no cover
         *,
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None,
         container: DependencyContainer = None,
         raise_on_missing: bool = True
         ) -> C: ...


@overload
def wire(*,  # noqa: E704  # pragma: no cover
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None,
         container: DependencyContainer = None,
         raise_on_missing: bool = True
         ) -> Callable[[C], C]: ...


def wire(class_: type = None,
         *,
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None,
         container: DependencyContainer = None,
         raise_on_missing: bool = True
         ) -> Union[Callable, type]:
    """Wire a class by injecting the dependencies in all specified methods.

    Injection arguments (dependencies, use_names, use_type_hints) are adapted
    for match the arguments of the method. Hence :code:`@inject` won't raise
    an error that it has too much dependencies.

    Args:
        class_: class to wire.
        methods: Name of the methods for which dependencies should be
            injected. Defaults to all defined methods.
        dependencies: Can be either a mapping of arguments name to their
            dependency, an iterable of dependencies or a function which returns
            the dependency given the arguments name. If an iterable is specified,
            the position of the arguments is used to determine their respective
            dependency. An argument may be skipped by using :code:`None` as a
            placeholder. The first argument is always ignored for methods (self)
            and class methods (cls).Type hints are overridden. Defaults to :code:`None`.
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to restrict this to those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints (annotations) should be
            used as the arguments dependency. An iterable of argument names may
            also be specified to restrict this to those. Any type hints from
            the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
            ...) are ignored. Defaults to :code:`True`.
        wire_super: If a method from a super-class needs to be wired, specify
            either a list of method names or :code:`True` to enable it for
            all methods. Defaults to :code:`False`, only methods defined in the
            class itself can be wired.
        container: :py:class:`~.core.container.DependencyContainer` from which
            the dependencies should be retrieved. Defaults to the global
            core if it is defined.
        raise_on_missing: Raise an error if a method does exist.
            Defaults to :code:`True`.

    Returns:
        Wired class or a decorator.

    """
    if not isinstance(methods, c_abc.Iterable):
        raise TypeError("methods must be either None or an iterable.")

    methods = set(methods)
    wire_super = _validate_wire_super(wire_super, methods)

    if not isinstance(raise_on_missing, bool):
        raise TypeError("raise_on_missing must be a boolean, "
                        "not a {!r}".format(type(raise_on_missing)))

    if isinstance(dependencies, c_abc.Iterable) \
            and not isinstance(dependencies, c_abc.Mapping):
        # convert to Tuple in case we cannot iterate more than once.
        dependencies = tuple(dependencies)

    def wire_methods(cls):
        if not inspect.isclass(cls):
            raise TypeError("Expecting a class, got a {}".format(type(cls)))

        for method_name in methods:
            method = _get_method(cls, method_name,
                                 with_super=method_name in wire_super)

            if method is None:
                if raise_on_missing:
                    raise TypeError("{!r} does not have a method "
                                    "named {!r}".format(cls, method_name))
                else:
                    continue  # pragma: no cover

            arguments = Arguments.from_callable(method)
            _dependencies = dependencies
            _use_names = use_names
            _use_type_hints = use_type_hints

            # Restrict injection parameters to those really needed by the method.
            if isinstance(dependencies, c_abc.Mapping):
                _dependencies = {
                    arg_name: dependency
                    for arg_name, dependency in dependencies.items()
                    if arg_name in arguments.without_self
                }
            elif isinstance(dependencies, c_abc.Iterable):
                _dependencies = dependencies[:len(arguments.without_self)]

            if isinstance(use_names, c_abc.Iterable):
                _use_names = [name
                              for name in use_names
                              if name in arguments]

            if isinstance(use_type_hints, c_abc.Iterable):
                _use_type_hints = [name
                                   for name in use_type_hints
                                   if name in arguments]

            injected_method = inject(method,
                                     arguments=arguments,
                                     dependencies=_dependencies,
                                     use_names=_use_names,
                                     use_type_hints=_use_type_hints,
                                     container=container)

            if injected_method is not method:  # If something has changed
                setattr(cls, method_name, injected_method)

        return cls

    return class_ and wire_methods(class_) or wire_methods


def _validate_wire_super(wire_super: Optional[Union[bool, Iterable[str]]],
                         methods: Set[str]) -> Set[str]:
    if wire_super is None:
        return set()

    if isinstance(wire_super, bool):
        return set(methods) if wire_super else set()

    if isinstance(wire_super, c_abc.Iterable):
        wire_super = set(wire_super)
        if not wire_super.issubset(methods):
            raise ValueError(
                "Method names {!r} are not specified "
                "not specified in methods".format(wire_super - methods)
            )
        return wire_super

    raise TypeError("wire_super must be either a boolean "
                    "or an iterable of method names.")


def _get_method(cls: type, name: str, with_super: bool) -> Optional[Callable]:
    if with_super:
        for c in cls.__mro__:
            wrapped = c.__dict__.get(name)
            if wrapped is not None:
                return wrapped
    else:
        return cls.__dict__.get(name)

    return None



from enum import Enum

import pytest

from antidote.core import DependencyContainer
from antidote.exceptions import DuplicateDependencyError, UndefinedContextError
from antidote.providers.indirect import IndirectProvider
from antidote.providers.factory import FactoryProvider


@pytest.fixture
def container():
    return DependencyContainer()


@pytest.fixture
def factory_provider(container):
    provider = FactoryProvider(container=container)
    container.register_provider(provider)
    return provider


@pytest.fixture
def interface_provider(container):
    provider = IndirectProvider(container=container)
    container.register_provider(provider)
    return provider


class IService:
    pass


class Service(IService):
    pass


class ServiceA(IService):
    pass


class ServiceB(IService):
    pass


class Profile(Enum):
    A = 'A'
    B = 'B'


def test_none(interface_provider: IndirectProvider):
    assert interface_provider.provide(IService) is None


def test_interface(container: DependencyContainer,
                   interface_provider: IndirectProvider,
                   factory_provider: FactoryProvider):
    factory_provider.register_class(Service)
    interface_provider.register(IService, Service)

    expected = container.get(Service)
    assert interface_provider.provide(IService).instance is expected
    assert interface_provider.provide(IService).singleton is True


def test_not_singleton_interface(container: DependencyContainer,
                                 interface_provider: IndirectProvider,
                                 factory_provider: FactoryProvider):
    factory_provider.register_class(Service, singleton=False)
    interface_provider.register(IService, Service)

    service = container.get(Service)
    assert interface_provider.provide(IService).instance is not service
    assert isinstance(interface_provider.provide(IService).instance, Service)
    assert interface_provider.provide(IService).singleton is False


def test_profile_instance(container: DependencyContainer,
                          interface_provider: IndirectProvider,
                          factory_provider: FactoryProvider):
    factory_provider.register_class(ServiceA)
    factory_provider.register_class(ServiceB, singleton=False)
    interface_provider.register(IService, ServiceA, Profile.A)
    interface_provider.register(IService, ServiceB, Profile.B)
    current_profile = Profile.A

    def get_profile():
        return current_profile

    factory_provider.register_factory(Profile, singleton=False, factory=get_profile)

    service_a = container.get(ServiceA)
    service_b = container.get(ServiceB)

    assert service_a is interface_provider.provide(IService).instance
    assert interface_provider.provide(IService).singleton is False

    current_profile = Profile.B
    assert service_b is not interface_provider.provide(IService).instance
    assert isinstance(interface_provider.provide(IService).instance, ServiceB)
    assert interface_provider.provide(IService).singleton is False


def test_singleton_profile_instance(container: DependencyContainer,
                                    interface_provider: IndirectProvider,
                                    factory_provider: FactoryProvider):
    factory_provider.register_class(ServiceA)
    factory_provider.register_class(ServiceB, singleton=False)
    interface_provider.register(IService, ServiceA, Profile.A)
    interface_provider.register(IService, ServiceB, Profile.B)

    service_a = container.get(ServiceA)

    container.update_singletons({Profile: Profile.A})
    assert service_a is interface_provider.provide(IService).instance
    assert interface_provider.provide(IService).singleton is True


def test_singleton_profile_instance_2(container: DependencyContainer,
                                      interface_provider: IndirectProvider,
                                      factory_provider: FactoryProvider):
    factory_provider.register_class(ServiceA)
    factory_provider.register_class(ServiceB, singleton=False)
    interface_provider.register(IService, ServiceA, Profile.A)
    interface_provider.register(IService, ServiceB, Profile.B)

    service_b = container.get(ServiceB)

    container.update_singletons({Profile: Profile.B})
    assert service_b is not interface_provider.provide(IService).instance
    assert isinstance(interface_provider.provide(IService).instance, ServiceB)
    assert interface_provider.provide(IService).singleton is False


def test_invalid_profile(interface_provider: IndirectProvider):
    with pytest.raises(TypeError):
        interface_provider.register(IService, Service, 1)


@pytest.mark.parametrize(
    'first,second',
    [
        ((ServiceA,), (ServiceB,)),
        ((ServiceA,), (ServiceB, Profile.B)),
        ((ServiceA, Profile.A), (ServiceB,)),
        ((ServiceA, Profile.A), (ServiceB, Profile.A)),
    ]
)
def test_duplicate(interface_provider: IndirectProvider, first: tuple, second: tuple):
    interface_provider.register(IService, *first)

    with pytest.raises(DuplicateDependencyError, match=str(IService)):
        interface_provider.register(IService, *second)


def test_undefined_context(container: DependencyContainer,
                           interface_provider: IndirectProvider,
                           factory_provider: FactoryProvider):
    factory_provider.register_class(ServiceA)
    factory_provider.register_class(ServiceB)
    interface_provider.register(IService, ServiceA, Profile.A)
    container.update_singletons({Profile: Profile.B})

    with pytest.raises(UndefinedContextError):
        interface_provider.provide(IService)



# cython: language_level=3
# cython: boundscheck=False, wraparound=False
# @formatter:off
from antidote._internal.stack cimport DependencyStack
# @formatter:on

cdef class DependencyInstance:
    cdef:
        readonly object instance
        readonly bint singleton

cdef class DependencyContainer:
    cdef:
        object __weakref__
        list _providers
        dict _type_to_provider
        dict _singletons
        DependencyStack _dependency_stack
        object _instantiation_lock

    cpdef object get(self, object dependency)
    cpdef DependencyInstance safe_provide(self, object dependency)
    cpdef DependencyInstance provide(self, object dependency)

cdef class DependencyProvider:
    cdef:
        public DependencyContainer _container

    cpdef DependencyInstance provide(self, object dependency)

cdef class Lazy:
    cdef:
        public object dependency



import pytest

from antidote.core import DependencyContainer
from antidote.exceptions import DuplicateDependencyError
from antidote.providers.factory import Build, FactoryProvider


class Service:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class AnotherService(Service):
    pass


@pytest.fixture()
def provider():
    container = DependencyContainer()
    provider = FactoryProvider(container=container)
    container.register_provider(provider)
    return provider


@pytest.mark.parametrize(
    'wrapped,kwargs',
    [
        (1, {'test': 1}),
        (Service, {'another': 'no'}),
        (Service, {'not_hashable': {'hey': 'hey'}})
    ]
)
def test_build_eq_hash(wrapped, kwargs):
    b = Build(wrapped, **kwargs)

    # does not fail
    hash(b)

    for f in (lambda e: e, hash):
        assert f(Build(wrapped, **kwargs)) == f(b)

    assert repr(wrapped) in repr(b)
    assert repr(kwargs) in repr(b)


@pytest.mark.parametrize(
    'args,kwargs',
    [
        [(), {}],
        [(1,), {}],
        [(), {'test': 1}],
    ]
)
def test_invalid_build(args: tuple, kwargs: dict):
    with pytest.raises(TypeError):
        Build(*args, **kwargs)


def test_simple(provider: FactoryProvider):
    provider.register_class(Service)

    dependency = provider.provide(Service)
    assert isinstance(dependency.instance, Service)
    assert repr(Service) in repr(provider)


def test_singleton(provider: FactoryProvider):
    provider.register_class(Service, singleton=True)
    provider.register_class(AnotherService, singleton=False)

    provide = provider.provide
    assert provide(Service).singleton is True
    assert provide(AnotherService).singleton is False


def test_takes_dependency(provider: FactoryProvider):
    provider.register_factory(factory=lambda cls: cls(), dependency=Service,
                              takes_dependency=True)

    assert isinstance(provider.provide(Service).instance, Service)
    assert provider.provide(AnotherService) is None


def test_build(provider: FactoryProvider):
    provider.register_class(Service)

    s = provider.provide(Build(Service, val=object)).instance
    assert isinstance(s, Service)
    assert dict(val=object) == s.kwargs

    provider.register_factory(AnotherService, factory=AnotherService,
                              takes_dependency=True)

    s = provider.provide(Build(AnotherService, val=object)).instance
    assert isinstance(s, AnotherService)
    assert (AnotherService,) == s.args
    assert dict(val=object) == s.kwargs


def test_non_singleton_factory(provider: FactoryProvider):
    def factory_builder():
        def factory(o=object()):
            return o

        return factory

    provider.register_factory('factory', factory=factory_builder, singleton=False)
    provider.register_providable_factory('service', factory_dependency='factory')

    service = provider.provide('service').instance
    assert provider.provide('service').instance is not service


def test_duplicate_error(provider: FactoryProvider):
    provider.register_class(Service)

    with pytest.raises(DuplicateDependencyError):
        provider.register_class(Service)

    with pytest.raises(DuplicateDependencyError):
        provider.register_factory(factory=lambda: Service(), dependency=Service)

    with pytest.raises(DuplicateDependencyError):
        provider.register_providable_factory(factory_dependency='dummy',
                                             dependency=Service)


@pytest.mark.parametrize(
    'kwargs',
    [dict(factory='test', dependency=Service),
     dict(factory=object(), dependency=Service)]
)
def test_invalid_type(provider: FactoryProvider, kwargs):
    with pytest.raises(TypeError):
        provider.register_factory(**kwargs)


@pytest.mark.parametrize('dependency', ['test', Service, object()])
def test_unknown_dependency(provider: FactoryProvider, dependency):
    assert provider.provide(dependency) is None



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from typing import Any, Callable, Dict, Hashable, Optional

# @formatter:off
from cpython.dict cimport PyDict_GetItem
from cpython.ref cimport PyObject

from antidote.core.container cimport (DependencyContainer, DependencyInstance,
                                     DependencyProvider)
from ..exceptions import DuplicateDependencyError
# @formatter:on


cdef class Build:
    """
    Custom Dependency wrapper used to pass arguments to the factory used to
    create the actual dependency.

    .. doctest::

        >>> from antidote import Build, register, world
        >>> @register
        ... class Dummy:
        ...     def __init__(self, name=None):
        ...         self.name = name
        >>> dummy = world.get(Build(Dummy, name='me'))
        >>> dummy.name
        'me'

    With no arguments, that is to say :code:`Build(x)`, it is equivalent to
    :code:`x` for the :py:class:`~.core.DependencyContainer`.
    """

    def __init__(self, dependency: Hashable, **kwargs):
        """
        Args:
            *args: The first argument is the dependency, all others are passed
                on to the factory.
            **kwargs: Passed on to the factory.
        """
        self.dependency = dependency  # type: Hashable
        self.kwargs = kwargs  # type: Dict

        if not self.kwargs:
            raise TypeError("Without additional arguments, Build must not be used.")

        try:
            # Try most precise hash first
            self._hash = hash((self.dependency, tuple(self.kwargs.items())))
        except TypeError:
            # If type error, return the best error-free hash possible
            self._hash = hash((self.dependency, tuple(self.kwargs.keys())))

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return "{}(id={!r}, kwargs={!r})".format(type(self).__name__,
                                                 self.dependency,
                                                 self.kwargs)

    __str__ = __repr__

    def __eq__(self, other):
        return isinstance(other, Build) \
               and (self.dependency is other.dependency
                    or self.dependency == other.dependency) \
               and self.kwargs == self.kwargs  # noqa

cdef class FactoryProvider(DependencyProvider):
    """
    Provider managing factories. Also used to register classes directly.
    """
    bound_dependency_types = (Build,)

    def __init__(self, DependencyContainer container):
        super().__init__(container)
        self._builders = dict()  # type: Dict[Any, Builder]

    def __repr__(self):
        return "{}(factories={!r})".format(type(self).__name__,
                                           tuple(self._builders.keys()))

    cpdef DependencyInstance provide(self, object dependency: Hashable):
        cdef:
            Builder builder
            Build build
            PyObject*ptr
            DependencyInstance f
            object instance
            object factory

        if isinstance(dependency, Build):
            build = <Build> dependency
            ptr = PyDict_GetItem(self._builders, build.dependency)
        else:
            ptr = PyDict_GetItem(self._builders, dependency)

        if ptr == NULL:
            return None

        builder = <Builder> ptr

        if builder.factory_dependency is not None:
            f = self._container.safe_provide(builder.factory_dependency)
            if f.singleton:
                builder.factory_dependency = None
                builder.factory = f.instance
            factory = f.instance
        else:
            factory = builder.factory

        if isinstance(dependency, Build):
            if builder.takes_dependency:
                instance = factory(build.dependency, **build.kwargs)
            else:
                instance = factory(**build.kwargs)
        else:
            if builder.takes_dependency:
                instance = factory(dependency)
            else:
                instance = factory()

        return DependencyInstance.__new__(DependencyInstance,
                                          instance,
                                          builder.singleton)

    def register_class(self, class_: type, singleton: bool = True):
        """
        Register a class which is both dependency and factory.

        Args:
            class_: dependency to register.
            singleton: Whether the dependency should be mark as singleton or
                not for the :py:class:`~..core.DependencyContainer`.
        """
        self.register_factory(dependency=class_, factory=class_,
                              singleton=singleton, takes_dependency=False)
        return class_

    def register_factory(self,
                         dependency: Hashable,
                         factory: Callable,
                         singleton: bool = True,
                         takes_dependency: bool = False):
        """
        Registers a factory for a dependency.

        Args:
            dependency: dependency to register.
            factory: Callable used to instantiate the dependency.
            singleton: Whether the dependency should be mark as singleton or
                not for the :py:class:`~..core.DependencyContainer`.
            takes_dependency: If True, the factory will be given the requested
                dependency as its first arguments. This allows re-using the
                same factory for different dependencies.
        """
        if dependency in self._builders:
            raise DuplicateDependencyError(dependency,
                                           self._builders[dependency])

        if callable(factory):
            self._builders[dependency] = Builder(singleton=singleton,
                                                 takes_dependency=takes_dependency,
                                                 factory=factory)
        else:
            raise TypeError("factory must be callable, not {!r}.".format(type(factory)))

    def register_providable_factory(self,
                                    dependency: Hashable,
                                    factory_dependency: Hashable,
                                    singleton: bool = True,
                                    takes_dependency: bool = False):
        """
        Registers a lazy factory (retrieved only at the first instantiation) for
        a dependency.

        Args:
            dependency: dependency to register.
            factory_dependency: Dependency used to retrieve.
            singleton: Whether the dependency should be mark as singleton or
                not for the :py:class:`~..core.DependencyContainer`.
            takes_dependency: If True, the factory will be given the requested
                dependency as its first arguments. This allows re-using the
                same factory for different dependencies.
        """
        if dependency in self._builders:
            raise DuplicateDependencyError(dependency,
                                           self._builders[dependency])

        self._builders[dependency] = Builder(singleton=singleton,
                                             takes_dependency=takes_dependency,
                                             factory_dependency=factory_dependency)

cdef class Builder:
    """
    Not part of the public API.

    Only used by the FactoryProvider to store information on how the factory
    has to be used.
    """
    cdef:
        bint singleton
        bint takes_dependency
        object factory
        object factory_dependency

    def __init__(self,
                 bint singleton,
                 bint takes_dependency,
                 factory: Optional[Callable] = None,
                 factory_dependency: Optional[Hashable] = None):
        assert factory is not None or factory_dependency is not None
        self.singleton = singleton
        self.takes_dependency = takes_dependency
        self.factory = factory
        self.factory_dependency = factory_dependency

    def __repr__(self):
        return ("{}(singleton={!r}, takes_dependency={!r}, factory={!r},"
                "factory_dependency={!r})").format(
            type(self).__name__,
            self.singleton,
            self.takes_dependency,
            self.factory,
            self.factory_dependency)



import itertools

import pytest

from antidote import factory, new_container, register, Tag, Tagged, TaggedDependencies


class Service:
    pass


class AnotherService:
    pass


class ServiceFactory:
    def __call__(self) -> Service:
        return Service()


def service_factory() -> Service:
    return Service()


def conf(key):
    return dict(test=object())[key]


@pytest.fixture()
def container():
    return new_container()


tag_tests = [
    [register, Service],
    [factory, ServiceFactory],
    [factory, service_factory],
]


def parametrize_tagging(tags):
    def decorator(test):
        @pytest.mark.parametrize(
            'wrapper,wrapped,tags',
            [
                pytest.param(wrapper, wrapped, tags_,
                             id="{}-{}-{}".format(
                                 wrapper.__name__,
                                 wrapped.__name__,
                                 "+".join(map(str, tags_))
                             ))
                for (wrapper, wrapped), tags_
                in itertools.product(tag_tests, tags)
            ]

        )
        def f(container, wrapper, wrapped, tags):
            if isinstance(wrapped, type):
                # helpers do modify the class, so a copy has to be made to
                # avoid any conflict between the tests.
                wrapped = type(wrapped.__name__,
                               wrapped.__bases__,
                               wrapped.__dict__.copy())
            wrapped = wrapper(container=container, tags=tags)(wrapped)

            if wrapper == register:
                dependency = wrapped
            else:
                dependency = Service

            return test(container, dependency=dependency, tags=tags)

        return f

    return decorator


@parametrize_tagging([
    [Tag('test')],
    [Tag('test'), Tag('test2')],
    [Tag('test'), 'test2'],
    ['test'],
])
def test_multi_tags(container, dependency, tags):
    for tag in tags:
        tag_name = tag if isinstance(tag, str) else tag.name
        tagged_dependencies = container.get(
            Tagged(tag_name))  # type: TaggedDependencies  # noqa

        assert 1 == len(tagged_dependencies)
        assert [container.get(dependency) == list(tagged_dependencies.instances())]
        assert [dependency] == list(tagged_dependencies.dependencies())

        bound_tag, = list(tagged_dependencies.tags())
        if isinstance(tag, Tag):
            assert tag is bound_tag
        else:
            assert isinstance(bound_tag, Tag)
            assert tag_name == bound_tag.name



from typing import cast

import pytest

from antidote import register
from antidote.core import DependencyContainer
from antidote.providers import FactoryProvider, TagProvider


@pytest.fixture()
def container():
    container = DependencyContainer()
    container.register_provider(FactoryProvider(container=container))
    container.register_provider(TagProvider(container=container))

    return container


def test_simple(container: DependencyContainer):
    @register(container=container)
    class Service:
        pass

    assert isinstance(container.get(Service), Service)
    # singleton by default
    assert container.get(Service) is container.get(Service)


def test_singleton(container: DependencyContainer):
    @register(container=container, singleton=True)
    class Singleton:
        pass

    assert container.get(Singleton) is container.get(Singleton)

    @register(container=container, singleton=False)
    class NoScope:
        pass

    assert container.get(NoScope) != container.get(NoScope)


@pytest.mark.parametrize(
    'factory',
    [
        lambda cls: cls(),
        'class_build',
        'static_build',
        None
    ]
)
def test_factory(container: DependencyContainer, factory):
    @register(container=container, factory=factory)
    class Service:
        @classmethod
        def class_build(cls):
            return cls()

        @staticmethod
        def static_build(cls):
            return cls()

    assert isinstance(container.get(Service), Service)

    @register(container=container, factory=factory)
    class SubService(Service):
        pass

    assert isinstance(container.get(SubService), SubService)


def test_factory_dependency(container: DependencyContainer):
    @register(container=container, factory_dependency='factory')
    class Service:
        pass

    container.update_singletons(dict(factory=lambda cls: dict(service=cls())))
    assert isinstance(container.get(Service), dict)
    assert isinstance(container.get(Service)['service'], Service)


@pytest.mark.parametrize('cls', ['test', object(), lambda: None])
def test_invalid_class(cls):
    with pytest.raises(TypeError):
        register(cls)


@pytest.mark.parametrize(
    'error,kwargs',
    [
        (TypeError, dict(factory=object())),
        (TypeError, dict(auto_wire=object())),
        (ValueError, dict(factory=lambda: None, factory_dependency=object())),
        (TypeError, dict(factory='method', auto_wire=False)),
    ]
)
def test_invalid_params(error, kwargs):
    with pytest.raises(error):
        @register(**kwargs)
        class Dummy:
            method = None


def test_invalid_factory_wiring():
    with pytest.raises(AttributeError):
        @register(factory='build')
        class Dummy:
            pass

    class NewDummy:
        @classmethod
        def build(cls):
            return cls()

    with pytest.raises(TypeError):
        @register(factory='build', wire_super=False)
        class Dummy2(NewDummy):
            pass



import threading
from typing import (Any, Callable, Dict, Hashable, Iterable, Iterator, List, Optional,
                    Union)

from .._internal.utils import SlotsReprMixin
from ..core import DependencyContainer, DependencyInstance, DependencyProvider
from ..exceptions import DuplicateTagError


class Tag(SlotsReprMixin):
    """
    Tags are a way to expose a dependency indirectly. Instead of explicitly
    defining a list of dependencies to retrieve, one can just mark those with
    tags and retrieve them. This helps decoupling dependencies, and may be used
    to add extensions to another service typically.

    The tag itself has a name (string) attribute, with which it is identified.
    One may add others attributes to pass additional information to the services
    retrieving it.

    .. doctest::

        >>> from antidote import Tag
        >>> t = Tag('dep', info=1)
        >>> t.info
        1

    """
    __slots__ = ('name', '_attrs')

    def __init__(self, name: str, **attrs):
        """
        Args:
            name: Name which identifies the tag.
            **attrs: Any other parameters will be accessible as an attribute.
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string")

        if len(name) == 0:
            raise ValueError("name must be a non empty string")

        self.name = name
        self._attrs = attrs

    def __str__(self):
        if not self._attrs:
            return "{}({!r})".format(type(self).__name__, self.name)
        return repr(self)

    def __repr__(self):
        return "{}(name={!r}, {})".format(
            type(self).__name__,
            self.name,
            ", ".join("{}={!r}".format(k, v) for k, v in self._attrs.items())
        )

    def __getattr__(self, item):
        return self._attrs.get(item)


class Tagged(SlotsReprMixin):
    """
    Custom dependency used to retrieve all dependencies tagged with by with the
    name.
    """
    __slots__ = ('name',)

    def __init__(self, name: str):
        """
        Args:
            name: Name of the tags which shall be retrieved.
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string")

        if len(name) == 0:
            raise ValueError("name must be a non empty string")

        self.name = name

    __str__ = SlotsReprMixin.__repr__  # type: Callable[['Tagged'], str]

    def __hash__(self):
        return object.__hash__(self)

    def __eq__(self, other):
        return object.__eq__(self, other)


class TagProvider(DependencyProvider):
    """
    Provider managing string tag. Tags allows one to retrieve a collection of
    dependencies marked by their creator.
    """
    bound_dependency_types = (Tagged,)

    def __init__(self, container: DependencyContainer):
        super().__init__(container)
        self._dependency_to_tag_by_tag_name = {}  # type: Dict[str, Dict[Hashable, Tag]]

    def __repr__(self):
        return "{}(tagged_dependencies={!r})".format(
            type(self).__name__,
            self._dependency_to_tag_by_tag_name
        )

    def provide(self, dependency: Hashable) -> Optional[DependencyInstance]:
        """
        Returns all dependencies matching the tag name specified with a
        :py:class:`~.dependency.Tagged`. For every other case, :obj:`None` is
        returned.

        Args:
            dependency: Only :py:class:`~.dependency.Tagged` is supported, all
                others are ignored.

        Returns:
            :py:class:`~.TaggedDependencies` wrapped in a
            :py:class:`~..core.Instance`.
        """
        if isinstance(dependency, Tagged):
            dependencies = []
            tags = []
            dependency_to_tag = self._dependency_to_tag_by_tag_name.get(dependency.name,
                                                                        {})

            for dependency_, tag in dependency_to_tag.items():
                dependencies.append(dependency_)
                tags.append(tag)

            return DependencyInstance(
                TaggedDependencies(
                    container=self._container,
                    dependencies=dependencies,
                    tags=tags
                ),
                # Whether the returned dependencies are singletons or not is
                # their decision to take.
                singleton=False
            )

        return None

    def register(self, dependency: Hashable, tags: Iterable[Union[str, Tag]]):
        """
        Mark a dependency with all the supplied tags. Raises
        :py:exc:`~.exceptions.DuplicateTagError` if the tag has already been
        used for this dependency.

        Args:
            dependency: dependency to register.
            tags: Iterable of tags which should be associated with the
                dependency
        """
        for tag in tags:
            if isinstance(tag, str):
                tag = Tag(tag)

            if not isinstance(tag, Tag):
                raise ValueError("Expecting tag of type Tag, not {}".format(type(tag)))

            if tag.name not in self._dependency_to_tag_by_tag_name:
                self._dependency_to_tag_by_tag_name[tag.name] = {dependency: tag}
            elif dependency not in self._dependency_to_tag_by_tag_name[tag.name]:
                self._dependency_to_tag_by_tag_name[tag.name][dependency] = tag
            else:
                raise DuplicateTagError(tag.name)


class TaggedDependencies:
    """
    Collection containing dependencies and their tags. Dependencies are lazily
    instantiated. This is thread-safe.

    Used by :py:class:`~.TagProvider` to return the dependencies matching a tag.
    """

    def __init__(self,
                 container: DependencyContainer,
                 dependencies: List[Hashable],
                 tags: List[Tag]):
        self._lock = threading.Lock()
        self._container = container
        self._dependencies = dependencies
        self._tags = tags
        self._instances = []  # type: List[Any]

    def __len__(self):
        return len(self._tags)

    def dependencies(self) -> Iterable[Hashable]:
        """
        Returns all the dependencies retrieved. This does not instantiate them.
        """
        return iter(self._dependencies)

    def tags(self) -> Iterable[Tag]:
        """
        Returns all the tags retrieved. This does not instantiate the
        dependencies.
        """
        return iter(self._tags)

    def instances(self) -> Iterator:
        """
        Returns the dependencies, in a stable order for multi-threaded
        environments.
        """
        i = -1
        for i, instance in enumerate(self._instances):
            yield instance

        i += 1
        while i < len(self):
            try:
                yield self._instances[i]
            except IndexError:
                with self._lock:
                    # If not other thread has already added the instance.
                    if i == len(self._instances):
                        self._instances.append(
                            self._container.get(self._dependencies[i])
                        )
                yield self._instances[i]
            i += 1



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from typing import Any, Dict, Iterable, Iterator, List, Union, Optional

# @formatter:off
from cpython.dict cimport PyDict_GetItem
from cpython.ref cimport PyObject
from fastrlock.rlock cimport create_fastrlock, lock_fastrlock, unlock_fastrlock

from antidote.core.container cimport (DependencyContainer, DependencyInstance,
                                      DependencyProvider)
# @formatter:on
from ..exceptions import DuplicateTagError

cdef class Tag:
    """
    Tags are a way to expose a dependency indirectly. Instead of explicitly
    defining a list of dependencies to retrieve, one can just mark those with
    tags and retrieve them. This helps decoupling dependencies, and may be used
    to add extensions to another service typically.

    The tag itself has a name (string) attribute, with which it is identified.
    One may add others attributes to pass additional information to the services
    retrieving it.

    .. doctest::

        >>> from antidote import Tag
        >>> t = Tag('dep', info=1)
        >>> t.info
        1

    """
    def __init__(self, str name: str, **attrs):
        """
        Args:
            name: Name which identifies the tag.
            **attrs: Any other parameters will be accessible as an attribute.
        """
        if len(name) == 0:
            raise ValueError("name must be a non empty string")

        self.name = name
        self._attrs = attrs

    def __repr__(self):
        return "{}(name={!r}, {})".format(
            type(self).__name__,
            self.name,
            ", ".join("{}={!r}".format(k, v) for k, v in self._attrs.items())
        )

    def __getattr__(self, item):
        return self._attrs.get(item)

cdef class Tagged:
    """
    Custom dependency used to retrieve all dependencies tagged with by with the
    name.
    """
    def __init__(self, str name):
        """
        Args:
            name: Name of the tags which shall be retrieved.
        """
        if len(name) == 0:
            raise ValueError("name must be a non empty string")

        self.name = name

    def __repr__(self):
        return "{}(name={!r})".format(type(self).__name__, self.name)

cdef class TagProvider(DependencyProvider):
    """
    Provider managing string tag. Tags allows one to retrieve a collection of
    dependencies marked by their creator.
    """
    bound_dependency_types = (Tagged,)

    def __init__(self, DependencyContainer container):
        super().__init__(container)
        self._dependency_to_tag_by_tag_name = {}  # type: Dict[str, Dict[Any, Tag]]

    def __repr__(self):
        return "{}(tagged_dependencies={!r})".format(
            type(self).__name__,
            self._dependency_to_tag_by_tag_name
        )

    cpdef DependencyInstance provide(self, dependency):
        """
        Returns all dependencies matching the tag name specified with a
        :py:class:`~.dependency.Tagged`. For every other case, :obj:`None` is
        returned.

        Args:
            dependency: Only :py:class:`~.dependency.Tagged` is supported, all
                others are ignored.

        Returns:
            :py:class:`~.TaggedDependencies` wrapped in a
            :py:class:`~..core.Instance`.
        """
        cdef:
            list dependencies
            list tags
            object dependency_
            Tagged tagged
            Tag tag
            PyObject*ptr

        if isinstance(dependency, Tagged):
            tagged = <Tagged> dependency
            dependencies = []
            tags = []
            ptr = PyDict_GetItem(self._dependency_to_tag_by_tag_name,
                                 tagged.name)

            if ptr != NULL:
                for dependency_, tag in (<dict> ptr).items():
                    dependencies.append(dependency_)
                    tags.append(tag)

            return DependencyInstance.__new__(
                DependencyInstance,
                TaggedDependencies.__new__(
                    TaggedDependencies,
                    container=self._container,
                    dependencies=dependencies,
                    tags=tags
                ),
                # Whether the returned dependencies are singletons or not is
                # their decision to take.
                singleton=False
            )

    def register(self, dependency, tags: Iterable[Union[str, Tag]]):
        """
        Mark a dependency with all the supplied tags. Raises
        :py:exc:`~.exceptions.DuplicateTagError` if the tag has already been
        used for this dependency.

        Args:
            dependency: dependency to register.
            tags: Iterable of tags which should be associated with the
                dependency
        """
        for tag in tags:
            if isinstance(tag, str):
                tag = Tag(tag)

            if not isinstance(tag, Tag):
                raise ValueError("Expecting tag of type Tag, not {}".format(type(tag)))

            if tag.name not in self._dependency_to_tag_by_tag_name:
                self._dependency_to_tag_by_tag_name[tag.name] = {dependency: tag}
            elif dependency not in self._dependency_to_tag_by_tag_name[tag.name]:
                self._dependency_to_tag_by_tag_name[tag.name][dependency] = tag
            else:
                raise DuplicateTagError(tag.name)

cdef class TaggedDependencies:
    """
    Collection containing dependencies and their tags. Dependencies are lazily
    instantiated. This is thread-safe.

    Used by :py:class:`~.TagProvider` to return the dependencies matching a tag.
    """
    def __cinit__(self,
                  DependencyContainer container,
                  list dependencies,
                  list tags):
        self._lock = create_fastrlock()
        self._container = container
        self._dependencies = dependencies  # type: List[Any]
        self._tags = tags  # type: List[Tag]
        self._instances = []  # type: List[Any]

    def __len__(self):
        return len(self._dependencies)

    def dependencies(self) -> Iterable[Any]:
        """
        Returns all the dependencies retrieved. This does not instantiate them.
        """
        return iter(self._dependencies)

    def tags(self) -> Iterable[Tag]:
        """
        Returns all the tags retrieved. This does not instantiate the
        dependencies.
        """
        return iter(self._tags)

    def instances(self) -> Iterator[Any]:
        """
        Returns the dependencies, in a stable order for multi-threaded
        environments.
        """
        cdef:
            ssize_t n = len(self._instances)
            ssize_t i = 0
            DependencyInstance dependency_instance
            object instance

        while i < len(self):
            if i < n:
                yield self._instances[i]
            else:
                lock_fastrlock(self._lock, -1, True)
                try:
                    # If not other thread has already added the instance.
                    if i < len(self._instances):
                        instance = self._instances[i]
                    else:
                        instance = self._container.get(self._dependencies[i])
                        self._instances.append(instance)
                    n += 1
                finally:
                    unlock_fastrlock(self._lock)

                yield instance
            i += 1



from antidote.core import DependencyInstance, DependencyProvider


class DummyProvider(DependencyProvider):
    singleton = True

    def __init__(self, data=None):
        self.data = data

    def provide(self, dependency):
        try:
            return DependencyInstance(self.data[dependency],
                                      singleton=self.singleton)
        except KeyError:
            pass


class DummyFactoryProvider(DependencyProvider):
    singleton = True

    def __init__(self, data=None):
        self.data = data or dict()

    def provide(self, dependency):
        try:
            return DependencyInstance(self.data[dependency](),
                                      singleton=self.singleton)
        except KeyError:
            pass



# cython: language_level=3
# cython: boundscheck=False, wraparound=False
from antidote.core.container cimport DependencyInstance, DependencyProvider

cdef class LazyCallProvider(DependencyProvider):
    cpdef DependencyInstance provide(self, object dependency)

cdef class LazyCall:
    cdef:
        object _func
        tuple _args
        dict _kwargs
        bint _singleton

cdef class LazyMethodCall:
    cdef:
        str _method_name
        bint _singleton
        tuple _args
        dict _kwargs
        str _key

    cdef object _call(self, object instance)



import pytest
from hypothesis import given, strategies as st

from antidote import Tag, Tagged
from antidote.core import DependencyContainer, DependencyInstance
from antidote.exceptions import DependencyNotFoundError, DuplicateTagError
from antidote.providers.tag import TaggedDependencies, TagProvider


class Service:
    pass


@pytest.fixture()
def provider():
    container = DependencyContainer()
    provider = TagProvider(container=container)
    container.register_provider(provider)
    return provider


def test_tag():
    t = Tag(name='test', val='x')

    assert 'test' == t.name
    assert 'x' == t.val
    assert t.anything is None
    assert "val='x'" in repr(t)
    assert "'test'" in repr(t)
    assert "val='x'" in str(t)
    assert "'test'" in str(t)

    t2 = Tag(name='test')
    assert "'test'" in str(t2)


@pytest.mark.parametrize('name,error', [('', ValueError),
                                        (object(), TypeError)])
def test_invalid_tag(name, error):
    with pytest.raises(error):
        Tag(name)


@given(st.builds(Tagged, name=st.sampled_from(['test', '987 jkh@è'])))
def test_tagged_eq_hash(tagged):
    # does not fail
    hash(tagged)

    for f in (lambda e: e, hash):
        assert f(Tagged(tagged.name)) != f(tagged)

    assert repr(tagged.name) in repr(tagged)


@pytest.mark.parametrize('name,error', [('', ValueError),
                                        (object(), TypeError)])
def test_invalid_tagged(name, error):
    with pytest.raises(error):
        Tagged(name)


def test_tagged_dependencies():
    tag1 = Tag('tag1')
    tag2 = Tag('tag2', dummy=True)
    c = DependencyContainer()

    t = TaggedDependencies(
        container=c,
        dependencies=['d', 'd2'],
        tags=[tag1, tag2]
    )

    assert {tag1, tag2} == set(t.tags())
    assert {'d', 'd2'} == set(t.dependencies())
    assert 2 == len(t)
    # instantiation from container
    c.update_singletons({'d': 'test', 'd2': 'test2'})
    assert {'test', 'test2'} == set(t.instances())
    # from cache
    c.update_singletons({'d': 'different', 'd2': 'different2'})
    assert {'test', 'test2'} == set(t.instances())


def test_tagged_dependencies_invalid_dependency():
    tag = Tag('tag1')
    c = DependencyContainer()

    t = TaggedDependencies(
        container=c,
        dependencies=['d'],
        tags=[tag]
    )
    assert ['d'] == list(t.dependencies())
    assert [tag] == list(t.tags())
    with pytest.raises(DependencyNotFoundError):
        list(t.instances())


def test_repr(provider: TagProvider):
    provider = TagProvider(DependencyContainer())

    x = object()
    provider.register(x, [Tag(name='tag')])

    assert str(x) in repr(provider)


def test_provide_tags(provider: TagProvider):
    container = provider._container
    container.update_singletons(dict(test=object(), test2=object()))
    custom_tag = Tag('tag2', error=True)
    provider.register('test', ['tag1', custom_tag])
    provider.register('test2', ['tag2'])

    result = provider.provide(Tagged('xxxxx'))
    assert isinstance(result, DependencyInstance)
    assert result.singleton is False
    assert 0 == len(result.instance)

    result = provider.provide(Tagged('tag1'))
    assert isinstance(result, DependencyInstance)
    assert result.singleton is False
    #
    tagged_dependencies = result.instance  # type: TaggedDependencies
    assert 1 == len(tagged_dependencies)
    assert ['test'] == list(tagged_dependencies.dependencies())
    assert ['tag1'] == [tag.name for tag in tagged_dependencies.tags()]
    assert [container.get('test') == list(tagged_dependencies.instances())]

    result = provider.provide(Tagged('tag2'))
    assert isinstance(result, DependencyInstance)
    assert result.singleton is False

    tagged_dependencies = result.instance  # type: TaggedDependencies
    assert 2 == len(tagged_dependencies)
    assert {'test', 'test2'} == set(tagged_dependencies.dependencies())
    tags = list(tagged_dependencies.tags())
    assert {'tag2', 'tag2'} == {tag.name for tag in tags}
    assert any(tag is custom_tag for tag in tags)
    instances = {container.get('test'), container.get('test2')}
    assert instances == set(tagged_dependencies.instances())


@pytest.mark.parametrize('tag', ['tag', Tag(name='tag')])
def test_duplicate_tag_error(provider: TagProvider, tag):
    provider.register('test', [Tag(name='tag')])
    with pytest.raises(DuplicateTagError):
        provider.register('test', tags=[tag])


def test_duplicate_tag_error_in_same_register(provider: TagProvider):
    with pytest.raises(DuplicateTagError):
        provider.register('test', tags=[Tag(name='tag'), 'tag'])


@pytest.mark.parametrize(
    'tags',
    [
        [object],
        [lambda _: False],
        ['test', object]
    ]
)
def test_invalid_register(provider: TagProvider, tags):
    with pytest.raises(ValueError):
        provider.register('test', tags)


@pytest.mark.parametrize('dependency', ['test', Service, object()])
def test_unknown_dependency(provider: TagProvider, dependency):
    assert provider.provide(dependency) is None



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from typing import (Any, Dict, Hashable, List, Mapping, Tuple)

# @formatter:off
cimport cython
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.ref cimport PyObject
from fastrlock.rlock cimport create_fastrlock, lock_fastrlock, unlock_fastrlock

from antidote._internal.stack cimport DependencyStack
# @formatter:on
from ..exceptions import (DependencyCycleError, DependencyInstantiationError,
                          DependencyNotFoundError)

@cython.freelist(32)
cdef class DependencyInstance:
    """
    Simple wrapper used by a :py:class:`~.core.DependencyProvider` when returning
    an instance of a dependency so it can specify in which scope the instance
    belongs to.
    """
    def __cinit__(self, object instance, bint singleton = False):
        self.instance = instance
        self.singleton = singleton

    def __repr__(self):
        return "{}(instance={!r}, singleton={!r})".format(type(self).__name__,
                                                          self.instance,
                                                          self.singleton)

cdef class DependencyContainer:
    """
    Instantiates the dependencies through the registered providers and handles
    their scope.
    """

    def __init__(self):
        self._providers = list()  # type: List[DependencyProvider]
        self._type_to_provider = dict()  # type: Dict[type, DependencyProvider]
        self._singletons = dict()  # type: Dict[Any, DependencyInstance]
        self._singletons[DependencyContainer] = DependencyInstance(self, True)
        self._dependency_stack = DependencyStack()
        self._instantiation_lock = create_fastrlock()

    def __str__(self):
        return "{}(providers={!r}, type_to_provider={!r})".format(
            type(self).__name__,
            self._providers,
            self._type_to_provider
        )

    def __repr__(self):
        return "{}(providers={!r}, type_to_provider={!r}, singletons={!r})".format(
            type(self).__name__,
            self._providers,
            self._type_to_provider,
            self._singletons
        )

    @property
    def providers(self):
        """ Returns a mapping of all the registered providers by their type. """
        return {type(p): p for p in self._providers}

    @property
    def singletons(self):
        """ Returns all the defined singletons """
        return self._singletons.copy()

    def register_provider(self, provider: Hashable):
        """
        Registers a provider, which can then be used to instantiate dependencies.

        Args:
            provider: Provider instance to be registered.

        """
        if not isinstance(provider, DependencyProvider):
            raise TypeError("provider must be a DependencyProvider, not a {!r}".format(
                type(provider)
            ))

        for bound_type in provider.bound_dependency_types:
            if bound_type in self._type_to_provider:
                raise RuntimeError(
                    "Cannot bind {!r} to provider, already bound to {!r}".format(
                        bound_type, self._type_to_provider[bound_type]
                    )
                )

        for bound_type in provider.bound_dependency_types:
            self._type_to_provider[bound_type] = provider

        self._providers.append(provider)

    def update_singletons(self, dependencies: Mapping):
        """
        Update the singletons.
        """
        lock_fastrlock(self._instantiation_lock, -1, True)
        self._singletons.update({
            k: DependencyInstance(v, singleton=True)
            for k, v in dependencies.items()
        })
        unlock_fastrlock(self._instantiation_lock)

    cpdef object get(self, object dependency: Hashable):
        """
        Returns an instance for the given dependency. All registered providers
        are called sequentially until one returns an instance.  If none is
        found, :py:exc:`~.exceptions.DependencyNotFoundError` is raised.

        Args:
            dependency: Passed on to the registered providers.

        Returns:
            instance for the given dependency
        """
        return self.safe_provide(dependency).instance

    cpdef DependencyInstance safe_provide(self, object dependency):
        cdef:
            DependencyInstance dependency_instance

        dependency_instance = self.provide(dependency)
        if dependency_instance is None:
            raise DependencyNotFoundError(dependency)
        return dependency_instance

    cpdef DependencyInstance provide(self, object dependency: Hashable):
        """
        Internal method which should not be directly called. Prefer
        :py:meth:`~.core.core.DependencyContainer.get`.
        It may be overridden in a subclass to customize how dependencies are
        instantiated.

        Used by the injection wrappers.
        """
        cdef:
            DependencyInstance dependency_instance = None
            DependencyProvider provider
            PyObject*ptr
            Exception e
            list stack

        ptr = PyDict_GetItem(self._singletons, dependency)
        if ptr != NULL:
            return <DependencyInstance> ptr

        lock_fastrlock(self._instantiation_lock, -1, True)

        ptr = PyDict_GetItem(self._singletons, dependency)
        if ptr != NULL:
            unlock_fastrlock(self._instantiation_lock)
            return <DependencyInstance> ptr

        if 1 != self._dependency_stack.push(dependency):
            stack = self._dependency_stack._stack.copy()
            unlock_fastrlock(self._instantiation_lock)
            stack.append(dependency)
            raise DependencyCycleError(stack)

        try:
            ptr = PyDict_GetItem(self._type_to_provider, type(dependency))
            if ptr != NULL:
                dependency_instance = (<DependencyProvider> ptr).provide(dependency)
            else:
                for provider in self._providers:
                    dependency_instance = provider.provide(dependency)
                    if dependency_instance is not None:
                        break

            if dependency_instance is not None:
                if dependency_instance.singleton:
                    PyDict_SetItem(self._singletons, dependency, dependency_instance)
                return dependency_instance

        except Exception as e:
            if isinstance(e, DependencyCycleError):
                raise
            raise DependencyInstantiationError(dependency) from e
        finally:
            self._dependency_stack.pop()
            unlock_fastrlock(self._instantiation_lock)

        return None

cdef class DependencyProvider:
    """
    Abstract base class for a Provider.

    Used by the :py:class:`~.core.DependencyContainer` to instantiate
    dependencies. Several are used in a cooperative manner : the first instance
    to be returned by one of them is used. Thus providers should ideally not
    overlap and handle only one kind of dependencies such as strings or tag.

    This should be used whenever one needs to introduce a new kind of dependency,
    or control how certain dependencies are instantiated.
    """
    bound_dependency_types = ()  # type: Tuple[type]

    def __init__(self, DependencyContainer container):
        self._container = container

    cpdef DependencyInstance provide(self, dependency: Hashable):
        """
        Method called by the :py:class:`~.core.DependencyContainer` when
        searching for a dependency.

        It is necessary to check quickly if the dependency can be provided or
        not, as :py:class:`~.core.DependencyContainer` will try its
        registered providers. A good practice is to subclass
        :py:class:`~.core.Dependency` so custom dependencies be differentiated.

        Args:
            dependency: The dependency to be provided by the provider.

        Returns:
            The requested instance wrapped in a :py:class:`~.core.Instance`
            if available or :py:obj:`None`.
        """
        raise NotImplementedError()
